from __future__ import annotations

import asyncio
import logging
import pathlib
import subprocess
import sys
import uuid
from datetime import datetime
from functools import partial
from typing import Optional, cast

import click
from more_itertools import chunked
from setproctitle import setproctitle

from ai.backend.cli.params import BoolExprType, OptionalType
from ai.backend.cli.types import ExitCode
from ai.backend.common import redis_helper as redis_helper
from ai.backend.common.cli import LazyGroup
from ai.backend.common.validators import TimeDuration
from ai.backend.logging import BraceStyleAdapter, LogLevel

from .context import CLIContext, redis_ctx

log = BraceStyleAdapter(logging.getLogger("ai.backend.manager.cli"))


@click.group(invoke_without_command=False, context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "-f",
    "--config-path",
    "--config",
    type=click.Path(
        file_okay=True,
        dir_okay=False,
        exists=True,
        path_type=pathlib.Path,
    ),
    default=None,
    help="The config file path. (default: ./manager.conf and /etc/backend.ai/manager.conf)",
)
@click.option(
    "--debug",
    is_flag=True,
    help="Set the logging level to DEBUG",
)
@click.option(
    "--log-level",
    type=click.Choice([*LogLevel], case_sensitive=False),
    default=LogLevel.NOTSET,
    help="Set the logging verbosity level",
)
@click.pass_context
def main(
    ctx: click.Context,
    log_level: LogLevel,
    debug: bool,
    config_path: Optional[pathlib.Path] = None,
) -> None:
    """
    Manager Administration CLI
    """
    setproctitle("backend.ai: manager.cli")
    if debug:
        log_level = LogLevel.DEBUG

    ctx.obj = ctx.with_resource(CLIContext(config_path=config_path, log_level=log_level))


@main.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.option(
    "--psql-container",
    "container_name",
    type=str,
    default=None,
    metavar="ID_OR_NAME",
    help=(
        "Open a postgres client shell using the psql executable "
        "shipped with the given postgres container. "
        'If not set or set as an empty string "", it will auto-detect '
        "the psql container from the halfstack. "
        'If set "-", it will use the host-provided psql executable. '
        "You may append additional arguments passed to the psql cli command. "
        "[default: auto-detect from halfstack]"
    ),
)
@click.option(
    "--psql-help",
    is_flag=True,
    help="Show the help text of the psql command instead of this dbshell command.",
)
@click.argument("psql_args", nargs=-1, type=click.UNPROCESSED)
@click.pass_obj
def dbshell(cli_ctx: CLIContext, container_name, psql_help, psql_args):
    """
    Run the database shell.

    All additional arguments and options except `--psql-container` and `--psql-help` are
    transparently forwarded to the psql command.
    For instance, you can use `-c` to execute a psql/SQL statement on the command line.

    Note that you do not have to specify connection-related options
    because the dbshell command fills out them from the manager configuration.
    """
    bootstrap_config = asyncio.run(cli_ctx.get_bootstrap_config())
    db_config = bootstrap_config.db
    if psql_help:
        psql_args = ["--help"]
    if not container_name:
        # Try to get the database container name of the halfstack
        candidate_container_names = subprocess.check_output(
            ["docker", "ps", "--format", "{{.Names}}", "--filter", "name=half-db"],
        )
        if not candidate_container_names:
            click.echo(
                "Could not find the halfstack postgres container. "
                "Please set the container name explicitly.",
                err=True,
            )
            sys.exit(ExitCode.FAILURE)
        container_name = None
        name_list = candidate_container_names.decode("utf-8").splitlines()
        for name in name_list:
            if "exporter" in name:
                continue  # Skip exporter containers
            container_name = name
            break
        if not container_name:
            click.echo(
                "Could not find the halfstack postgres container. "
                "Please set the container name explicitly.",
                err=True,
            )
            sys.exit(ExitCode.FAILURE)
    elif container_name == "-":
        # Use the host-provided psql command
        cmd = [
            "psql",
            (f"postgres://{db_config.user}:{db_config.password}@{db_config.addr}/{db_config.name}"),
            *psql_args,
        ]
        subprocess.run(cmd)
        return
    # Use the container to start the psql client command
    log.info(f"using the db container {container_name} ...")
    cmd = [
        "docker",
        "exec",
        "-i",
        "-t",
        container_name,
        "psql",
        "-U",
        db_config.user,
        "-d",
        db_config.name,
        *psql_args,
    ]
    subprocess.run(cmd)


@main.command()
@click.pass_obj
def generate_api_keypair(cli_ctx: CLIContext) -> None:
    """
    Generate a manager API keypair and print it out to stdout.
    """
    from ..models.keypair import generate_keypair as _gen_keypair

    log.info("Generating a manager API keypair...")
    ak, sk = _gen_keypair()
    print(f"Access Key: {ak} ({len(ak)} bytes)")
    print(f"Secret Key: {sk} ({len(sk)} bytes)")


@main.command()
@click.argument(
    "dst_dir",
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        exists=True,
        writable=True,
        path_type=pathlib.Path,
    ),
)
@click.argument(
    "name",
    type=str,
)
@click.pass_obj
def generate_rpc_keypair(cli_ctx: CLIContext, dst_dir: pathlib.Path, name: str) -> None:
    """
    Generate a ZeroMQ CURVE keypair for use in manager-agent RPC and print it out to stdout.

    \b
    DST_DIR: The target directory to store .key/.key_secret files
    NAME: The name of generated key files
    """
    from zmq.auth.certs import create_certificates, load_certificate

    log.info("Generating a RPC keypair...")
    public_key_path, secret_key_path = create_certificates(dst_dir, name)
    public_key, secret_key = load_certificate(secret_key_path)
    assert secret_key is not None
    print(f"Public Key: {public_key.decode('ascii')} (stored at {public_key_path})")
    print(f"Secret Key: {secret_key.decode('ascii')} (stored at {secret_key_path})")


@main.command()
@click.option(
    "-r",
    "--retention",
    type=str,
    default="1yr",
    help="The retention limit. e.g., 20d, 1mo, 6mo, 1yr",
)
@click.option(
    "-v",
    "--vacuum-full",
    type=OptionalType(BoolExprType),
    default=False,
    help=(
        "Reclaim storage occupied by dead tuples."
        "If not set or set False, it will run VACUUM without FULL. "
        "If set True, it will run VACUUM FULL. "
        "When VACUUM FULL is being processed, the database is locked. "
        "[default: False]"
    ),
)
@click.pass_obj
def clear_history(cli_ctx: CLIContext, retention, vacuum_full) -> None:
    """
    Delete old records from the kernels, error_logs tables and
    invoke the PostgreSQL's vaccuum operation to clear up the actual disk space.
    """
    import sqlalchemy as sa
    from redis.asyncio import Redis
    from redis.asyncio.client import Pipeline

    from ai.backend.manager.models import SessionRow, error_logs, kernels
    from ai.backend.manager.models.utils import connect_database, vacuum_db

    today = datetime.now()
    duration = TimeDuration()
    expiration_date = today - duration.check_and_return(retention)
    bootstrap_config = asyncio.run(cli_ctx.get_bootstrap_config())

    async def _clear_redis_history():
        try:
            async with connect_database(bootstrap_config.db) as db:
                async with db.begin_readonly() as conn:
                    query = (
                        sa.select([kernels.c.id])
                        .select_from(kernels)
                        .where(
                            (kernels.c.terminated_at < expiration_date),
                        )
                    )
                    result = await conn.execute(query)
                    target_kernels = [str(x["id"]) for x in result.all()]

            delete_count = 0
            async with redis_ctx(cli_ctx) as redis_conn_set:

                async def _build_pipe(
                    r: Redis,
                    kernel_ids: list[str],
                ) -> Pipeline:
                    pipe = r.pipeline(transaction=False)
                    await pipe.delete(*kernel_ids)
                    return pipe

                if len(target_kernels) > 0:
                    # Apply chunking to avoid excessive length of command params
                    # and indefinite blocking of the Redis server.
                    for kernel_ids in chunked(target_kernels, 32):
                        results = await redis_helper.execute(
                            redis_conn_set.stat,
                            partial(_build_pipe, kernel_ids=kernel_ids),
                        )
                    # Each DEL command returns the number of keys deleted.
                    delete_count += sum(results)
                    log.info(
                        "Cleaned up {:,} redis statistics records older than {:}.",
                        delete_count,
                        expiration_date,
                    )

                # Sync and compact the persistent database of Redis
                redis_config = await redis_helper.execute(
                    redis_conn_set.stat,
                    lambda r: r.config_get("appendonly"),
                )
                if redis_config["appendonly"] == "yes":
                    await redis_helper.execute(
                        redis_conn_set.stat,
                        lambda r: r.bgrewriteaof(),
                    )
                    log.info("Issued BGREWRITEAOF to the Redis database.")
                else:
                    await redis_helper.execute(
                        redis_conn_set.stat,
                        lambda r: r.execute_command("BGSAVE SCHEDULE"),
                    )
                    log.info("Issued BGSAVE to the Redis database.")
        except Exception:
            log.exception("Unexpected error while cleaning up redis history")

    async def _clear_terminated_sessions():
        async with connect_database(bootstrap_config.db, isolation_level="AUTOCOMMIT") as db:
            async with db.begin() as conn:
                log.info("Deleting old records...")
                result = (
                    await conn.scalars(
                        sa.select(SessionRow.id).where(SessionRow.terminated_at < expiration_date)
                    )
                ).all()
                session_ids = cast(list[uuid.UUID], result)
                if session_ids:
                    await conn.execute(
                        sa.delete(kernels).where(kernels.c.session_id.in_(session_ids))
                    )
                    await conn.execute(sa.delete(SessionRow).where(SessionRow.id.in_(session_ids)))

                curs = await conn.execute(sa.select([sa.func.count()]).select_from(SessionRow))
                if ret := curs.fetchone():
                    table_size = ret[0]
                    log.info(
                        "The number of rows of the `sessions` tables after cleanup: {}",
                        table_size,
                    )
        log.info(
            "Cleaned up {:,} database records older than {}.",
            len(session_ids),
            expiration_date,
        )

    async def _clear_old_error_logs():
        async with connect_database(bootstrap_config.db, isolation_level="AUTOCOMMIT") as db:
            async with db.begin() as conn:
                log.info("Deleting old error logs...")
                result = await conn.execute(
                    sa.delete(error_logs).where(error_logs.c.created_at < expiration_date),
                )
                deleted_count = result.rowcount

        log.info(
            "Cleaned up {:,} error log records older than {}.",
            deleted_count,
            expiration_date,
        )

    asyncio.run(_clear_redis_history())
    asyncio.run(_clear_terminated_sessions())
    asyncio.run(_clear_old_error_logs())
    asyncio.run(vacuum_db(bootstrap_config, vacuum_full))


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.dbschema:cli")
def schema():
    """Command set for managing the database schema."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.etcd:cli")
def etcd():
    """Command set for putting/getting data to/from etcd."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.fixture:cli")
def fixture():
    """Command set for managing fixtures."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.api:cli")
def api():
    """Command set for API schema inspection and manipulation."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.gql:cli")
def gql():
    """Command set for GraphQL schema."""
    # Deprecated in favor of "api" but kept for backward compatibility


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.image:cli")
def image():
    """Command set for managing images."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.redis:cli")
def redis():
    """Command set for Redis related operations."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.agent:cli")
def agent():
    """Command set for agent related operations."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.scheduler:cli")
def scheduler():
    """Command set for scheduler related operations."""


@main.group(cls=LazyGroup, import_name="ai.backend.manager.cli.config:cli")
def config():
    """Command set for configuration management."""


if __name__ == "__main__":
    main()



from __future__ import annotations

import logging
from http import HTTPStatus
from typing import TYPE_CHECKING, Tuple

import aiohttp_cors
from aiohttp import web

from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.openapi import generate_openapi

from ..errors.common import GenericForbidden
from .auth import auth_required
from .types import CORSOptions, Iterable, WebMiddleware

if TYPE_CHECKING:
    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

OPENAPI_HTML = """
<!DOCTYPE html>
<html>
  <head>
    <title>Backend.AI REST API Reference</title>
    <meta charset="utf-8"/>
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <style>
      body {
        margin: 0;
        padding: 0;
      }
    </style>
  </head>
  <body>
    <redoc spec-url="openapi/spec.json"></redoc>
    <script src="../static/vendor/spec-viewer.js"></script>
  </body>
</html>
"""


GRAPHIQL_HTML = """
<html>
  <head>
    <title>Backend.AI GraphQL API Reference</title>
	<meta charset="UTF-8">
    <link href="../static/vendor/graphiql.min.css" rel="stylesheet" />
  </head>
  <body style="margin: 0;">
    <div id="graphiql" style="height: 100vh;"></div>

    <script src="../static/vendor/react.production.min.js"
    ></script>
    <script src="../static/vendor/react-dom.production.min.js"
    ></script>
    <script src="../static/vendor/graphiql.min.js"
    ></script>

    <script>
      const fetcher = GraphiQL.createFetcher({ url: '../admin/gql' });

      ReactDOM.render(
        React.createElement(GraphiQL, { fetcher: fetcher }),
        document.getElementById('graphiql'),
      );
    </script>
  </body>
</html>
"""


GRAPHIQL_V2_HTML = """
<html>
  <head>
    <title>Backend.AI GraphQL V2 API Reference</title>
	<meta charset="UTF-8">
    <link href="../../static/vendor/graphiql.min.css" rel="stylesheet" />
  </head>
  <body style="margin: 0;">
    <div id="graphiql" style="height: 100vh;"></div>

    <script src="../../static/vendor/react.production.min.js"
    ></script>
    <script src="../../static/vendor/react-dom.production.min.js"
    ></script>
    <script src="../../static/vendor/graphiql.min.js"
    ></script>

    <script>
      const fetcher = GraphiQL.createFetcher({ url: '../../admin/gql/strawberry' });

      ReactDOM.render(
        React.createElement(GraphiQL, { fetcher: fetcher }),
        document.getElementById('graphiql'),
      );
    </script>
  </body>
</html>
"""


@auth_required
async def render_graphiql_graphene_html(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    if not root_ctx.config_provider.config.api.allow_graphql_schema_introspection:
        raise GenericForbidden

    return web.Response(
        body=GRAPHIQL_HTML,
        status=HTTPStatus.OK,
        content_type="text/html",
    )


async def render_graphiql_strawberry_html(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    if not root_ctx.config_provider.config.api.allow_graphql_schema_introspection:
        raise GenericForbidden

    return web.Response(
        body=GRAPHIQL_V2_HTML,
        status=HTTPStatus.OK,
        content_type="text/html",
    )


@auth_required
async def render_openapi_html(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    if not root_ctx.config_provider.config.api.allow_openapi_schema_introspection:
        raise GenericForbidden

    return web.Response(
        body=OPENAPI_HTML,
        status=HTTPStatus.OK,
        content_type="text/html",
    )


@auth_required
async def generate_openapi_spec(request: web.Request) -> web.Response:
    root_ctx: RootContext = request.app["_root.context"]
    if not root_ctx.config_provider.config.api.allow_openapi_schema_introspection:
        raise GenericForbidden

    return web.json_response(generate_openapi(request.app["_root_app"]._subapps))


async def init(app: web.Application) -> None:
    root_ctx: RootContext = app["_root.context"]
    if root_ctx.config_provider.config.api.allow_openapi_schema_introspection:
        log.warning(
            "OpenAPI schema introspection is enabled. "
            "It is strongly advised to disable this in production setups."
        )


def create_app(
    default_cors_options: CORSOptions,
) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["api_versions"] = (4, 5)
    app["prefix"] = "spec"
    app.on_startup.append(init)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route("GET", "/graphiql", render_graphiql_graphene_html))
    cors.add(app.router.add_route("GET", "/graphiql/strawberry", render_graphiql_strawberry_html))
    cors.add(app.router.add_route("GET", "/openapi", render_openapi_html))
    cors.add(app.router.add_route("GET", "/openapi/spec.json", generate_openapi_spec))

    return app, []



from typing import override

from ai.backend.common.contexts.user import current_user
from ai.backend.manager.actions.action import BaseActionTriggerMeta
from ai.backend.manager.actions.action.single_entity import BaseSingleEntityAction
from ai.backend.manager.data.permission.id import ObjectId
from ai.backend.manager.data.permission.role import SingleEntityPermissionCheckInput
from ai.backend.manager.data.permission.types import EntityType
from ai.backend.manager.errors.user import UserNotFound
from ai.backend.manager.repositories.permission_controller.repository import (
    PermissionControllerRepository,
)

from ...validator.single_entity import SingleEntityActionValidator


class SingleEntityActionRBACValidator(SingleEntityActionValidator):
    def __init__(
        self,
        repository: PermissionControllerRepository,
    ) -> None:
        self._repository = repository

    @override
    async def validate(self, action: BaseSingleEntityAction, meta: BaseActionTriggerMeta) -> None:
        entity_type = EntityType(action.entity_type())
        entity_id = action.target_entity_id()
        user = current_user()
        if user is None:
            raise UserNotFound("User not found in context")

        await self._repository.check_permission_of_entity(
            SingleEntityPermissionCheckInput(
                user_id=user.user_id,
                operation=action.permission_operation_type(),
                target_object_id=ObjectId(
                    entity_type=entity_type,
                    entity_id=entity_id,
                ),
            )
        )



import logging
from io import BytesIO

import sqlalchemy as sa

from ai.backend.common.clients.valkey_client.valkey_container_log.client import (
    ValkeyContainerLogClient,
)
from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.clients.valkey_client.valkey_stream.client import ValkeyStreamClient
from ai.backend.common.events.event_types.kernel.anycast import (
    DoSyncKernelLogsEvent,
    KernelCancelledAnycastEvent,
    KernelCreatingAnycastEvent,
    KernelHeartbeatEvent,
    KernelPreparingAnycastEvent,
    KernelPullingAnycastEvent,
    KernelStartedAnycastEvent,
    KernelTerminatedAnycastEvent,
    KernelTerminatingAnycastEvent,
)
from ai.backend.common.types import (
    AgentId,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.registry import AgentRegistry
from ai.backend.manager.sokovan.scheduler.coordinator import ScheduleCoordinator

from ...models.kernel import kernels
from ...models.utils import (
    ExtendedAsyncSAEngine,
    execute_with_retry,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class KernelEventHandler:
    _valkey_container_log: ValkeyContainerLogClient
    _valkey_stat: ValkeyStatClient
    _valkey_stream: ValkeyStreamClient
    _registry: AgentRegistry
    _db: ExtendedAsyncSAEngine
    _schedule_coordinator: ScheduleCoordinator
    _use_sokovan: bool

    def __init__(
        self,
        valkey_container_log: ValkeyContainerLogClient,
        valkey_stat: ValkeyStatClient,
        valkey_stream: ValkeyStreamClient,
        registry: AgentRegistry,
        db: ExtendedAsyncSAEngine,
        schedule_coordinator: ScheduleCoordinator,
        use_sokovan: bool = False,
    ) -> None:
        self._valkey_container_log = valkey_container_log
        self._valkey_stat = valkey_stat
        self._valkey_stream = valkey_stream
        self._registry = registry
        self._db = db
        self._schedule_coordinator = schedule_coordinator
        self._use_sokovan = use_sokovan

    async def handle_kernel_log(
        self,
        context: None,
        source: AgentId,
        event: DoSyncKernelLogsEvent,
    ) -> None:
        # The log data is at most 10 MiB.
        log_buffer = BytesIO()
        try:
            list_size = await self._valkey_container_log.container_log_len(
                container_id=event.container_id
            )
            if list_size is None:
                # The log data is expired due to a very slow event delivery.
                # (should never happen!)
                log.warning(
                    "tried to store console logs for cid:{}, but the data is expired",
                    event.container_id,
                )
                return
            for _ in range(list_size):
                # Read chunk-by-chunk to allow interleaving with other Redis operations.
                chunks = await self._valkey_container_log.pop_container_logs(
                    container_id=event.container_id
                )
                if chunks is None:  # maybe missing
                    log_buffer.write(b"(container log unavailable)\n")
                    break
                for chunk in chunks:
                    log_buffer.write(chunk.get_content())
            try:
                log_data = log_buffer.getvalue()

                async def _update_log() -> None:
                    async with self._db.begin() as conn:
                        update_query = (
                            sa.update(kernels)
                            .values(container_log=log_data)
                            .where(kernels.c.id == event.kernel_id)
                        )
                        await conn.execute(update_query)

                await execute_with_retry(_update_log)
            finally:
                # Clear the log data from Redis when done.
                await self._valkey_container_log.clear_container_logs(
                    container_id=event.container_id
                )
        except Exception:
            # skip all exception in handle_kernel_log
            pass
        finally:
            log_buffer.close()

    async def handle_kernel_creation_lifecycle(
        self,
        context: None,
        source: AgentId,
        event: (
            KernelPreparingAnycastEvent
            | KernelPullingAnycastEvent
            | KernelCreatingAnycastEvent
            | KernelStartedAnycastEvent
            | KernelCancelledAnycastEvent
        ),
    ) -> None:
        """
        Update the database and perform post_create_kernel() upon
        the events for each step of kernel creation.

        To avoid race condition between consumer and subscriber event handlers,
        we only have this handler to subscribe all kernel creation events,
        but distinguish which one to process using a unique creation_id
        generated when initiating the create_kernels() agent RPC call.
        """
        log.debug(
            "handle_kernel_creation_lifecycle: ev:{} k:{}",
            event.event_name(),
            event.kernel_id,
        )
        match event:
            case KernelPreparingAnycastEvent():
                # State transition is done by the DoPrepareEvent handler inside the scheduler-distpacher object.
                pass
            case KernelPullingAnycastEvent(kernel_id, session_id, reason=reason):
                async with self._db.connect() as db_conn:
                    await self._registry.mark_kernel_pulling(db_conn, kernel_id, session_id, reason)
            case KernelCreatingAnycastEvent(kernel_id, session_id, reason=reason):
                async with self._db.connect() as db_conn:
                    await self._registry.mark_kernel_creating(
                        db_conn, kernel_id, session_id, reason
                    )
            case KernelStartedAnycastEvent(
                kernel_id, session_id, reason=reason, creation_info=creation_info
            ):
                async with self._db.connect() as db_conn:
                    await self._registry.mark_kernel_running(
                        db_conn, kernel_id, session_id, reason, creation_info
                    )
            case KernelCancelledAnycastEvent():
                log.warning(f"Kernel cancelled, {event.reason = }")

    async def handle_kernel_preparing(
        self,
        context: None,
        source: AgentId,
        event: KernelPreparingAnycastEvent,
    ) -> None:
        log.info(
            "handle_kernel_preparing: ev:{} k:{}",
            event.event_name(),
            event.kernel_id,
        )

        if self._use_sokovan:
            # Use Sokovan coordinator's kernel handlers
            await self._schedule_coordinator.handle_kernel_preparing(event)

    async def handle_kernel_pulling(
        self,
        context: None,
        source: AgentId,
        event: KernelPullingAnycastEvent,
    ) -> None:
        log.info(
            "handle_kernel_pulling: ev:{} k:{}",
            event.event_name(),
            event.kernel_id,
        )

        if self._use_sokovan:
            # Use Sokovan coordinator's kernel handlers
            await self._schedule_coordinator.handle_kernel_pulling(event)
        else:
            # Use legacy registry method
            async with self._db.connect() as db_conn:
                await self._registry.mark_kernel_pulling(
                    db_conn, event.kernel_id, event.session_id, event.reason
                )

    async def handle_kernel_creating(
        self,
        context: None,
        source: AgentId,
        event: KernelCreatingAnycastEvent,
    ) -> None:
        log.info(
            "handle_kernel_creating: ev:{} k:{}",
            event.event_name(),
            event.kernel_id,
        )

        if self._use_sokovan:
            # Use Sokovan coordinator's kernel handlers
            await self._schedule_coordinator.handle_kernel_creating(event)
        else:
            # Use legacy registry method
            async with self._db.connect() as db_conn:
                await self._registry.mark_kernel_creating(
                    db_conn, event.kernel_id, event.session_id, event.reason
                )

    async def handle_kernel_started(
        self,
        context: None,
        source: AgentId,
        event: KernelStartedAnycastEvent,
    ) -> None:
        log.info(
            "handle_kernel_started: ev:{} k:{}",
            event.event_name(),
            event.kernel_id,
        )

        if self._use_sokovan:
            # Use Sokovan coordinator's kernel handlers
            await self._schedule_coordinator.handle_kernel_running(event)
        else:
            # Use legacy registry method
            async with self._db.connect() as db_conn:
                await self._registry.mark_kernel_running(
                    db_conn, event.kernel_id, event.session_id, event.reason, event.creation_info
                )

    async def handle_kernel_cancelled(
        self,
        context: None,
        source: AgentId,
        event: KernelCancelledAnycastEvent,
    ) -> None:
        log.info(
            "handle_kernel_cancelled: ev:{} k:{}",
            event.event_name(),
            event.kernel_id,
        )

        if self._use_sokovan:
            # Use Sokovan coordinator's kernel handlers
            await self._schedule_coordinator.handle_kernel_cancelled(event)
        else:
            # Legacy code doesn't handle cancelled state
            log.warning(f"Kernel cancelled, {event.reason = }")

    async def handle_kernel_terminating(
        self,
        context: None,
        source: AgentId,
        event: KernelTerminatingAnycastEvent,
    ) -> None:
        # `destroy_kernel()` has already changed the kernel status to "TERMINATING".
        # No additional handling needed for terminating state
        pass

    async def handle_kernel_terminated(
        self,
        context: None,
        source: AgentId,
        event: KernelTerminatedAnycastEvent,
    ) -> None:
        if self._use_sokovan:
            # Use Sokovan coordinator's kernel handlers
            await self._schedule_coordinator.handle_kernel_terminated(event)
        else:
            # Use legacy registry method
            async with self._db.connect() as db_conn:
                await self._registry.mark_kernel_terminated(
                    db_conn, event.kernel_id, event.session_id, event.reason, event.exit_code
                )

    async def handle_kernel_heartbeat(
        self,
        context: None,
        source: AgentId,
        event: KernelHeartbeatEvent,
    ) -> None:
        if self._use_sokovan:
            # Use Sokovan coordinator's kernel handlers
            await self._schedule_coordinator.handle_kernel_heartbeat(event)
        else:
            # Use legacy registry method
            await self._registry.mark_kernel_heartbeat(event.kernel_id)



"""remove foreign keys on endpoints

Revision ID: 70e0533aa49a
Revises: 9fc0e92ea510
Create Date: 2025-03-05 14:10:39.897579

"""

from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "70e0533aa49a"
down_revision = "9fc0e92ea510"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "endpoint_tokens", "session_owner", existing_type=postgresql.UUID(), nullable=False
    )
    op.drop_constraint(
        "fk_endpoint_tokens_session_owner_users", "endpoint_tokens", type_="foreignkey"
    )
    op.drop_constraint(
        "fk_endpoint_tokens_endpoint_endpoints", "endpoint_tokens", type_="foreignkey"
    )
    op.drop_constraint("fk_endpoints_session_owner_users", "endpoints", type_="foreignkey")
    op.drop_constraint("fk_endpoints_created_user_users", "endpoints", type_="foreignkey")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_foreign_key(
        "fk_endpoints_created_user_users",
        "endpoints",
        "users",
        ["created_user"],
        ["uuid"],
        ondelete="RESTRICT",
    )
    op.create_foreign_key(
        "fk_endpoints_session_owner_users",
        "endpoints",
        "users",
        ["session_owner"],
        ["uuid"],
        ondelete="RESTRICT",
    )
    op.create_foreign_key(
        "fk_endpoint_tokens_endpoint_endpoints",
        "endpoint_tokens",
        "endpoints",
        ["endpoint"],
        ["id"],
        ondelete="SET NULL",
    )
    op.create_foreign_key(
        "fk_endpoint_tokens_session_owner_users",
        "endpoint_tokens",
        "users",
        ["session_owner"],
        ["uuid"],
    )
    op.alter_column(
        "endpoint_tokens", "session_owner", existing_type=postgresql.UUID(), nullable=True
    )
    # ### end Alembic commands ###



"""change permission table names

Revision ID: 643deb439458
Revises: ed84197be4fe
Create Date: 2025-07-11 22:10:52.126619

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "643deb439458"
down_revision = "ed84197be4fe"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.rename_table("role_permissions", "scope_permissions")
    op.drop_constraint("pk_role_permissions", "scope_permissions", type_="primary")
    op.create_primary_key("pk_scope_permissions", "scope_permissions", ["id"])

    op.rename_table("resource_permissions", "object_permissions")
    op.drop_constraint("pk_resource_permissions", "object_permissions", type_="primary")
    op.create_primary_key("pk_object_permissions", "object_permissions", ["id"])

    op.drop_column("user_roles", "state")
    op.drop_column("user_roles", "expires_at")
    op.drop_column("user_roles", "deleted_at")
    op.alter_column("roles", "state", new_column_name="status")
    op.add_column(
        "scope_permissions",
        sa.Column(
            "status",
            sa.VARCHAR(length=16),
            nullable=False,
            server_default="active",
        ),
    )
    op.add_column(
        "object_permissions",
        sa.Column(
            "status",
            sa.VARCHAR(length=16),
            nullable=False,
            server_default="active",
        ),
    )

    # Create indexes and constraints
    op.create_index(
        "ix_role_id_entity_id",
        "object_permissions",
        ["status", "role_id", "entity_id"],
        unique=False,
    )
    op.create_unique_constraint(
        "uq_object_permissions_entity_id_operation",
        "object_permissions",
        ["entity_id", "operation"],
    )
    op.create_index("ix_id_status", "roles", ["id", "status"], unique=False)
    op.create_index(
        "ix_role_id_entity_type_scope_id",
        "scope_permissions",
        ["status", "role_id", "entity_type", "scope_id"],
        unique=False,
    )
    op.create_unique_constraint(
        "uq_scope_permissions_entity_operation_scope_id",
        "scope_permissions",
        ["entity_type", "operation", "scope_id"],
    )
    op.create_unique_constraint("uq_user_id_role_id", "user_roles", ["user_id", "role_id"])


def downgrade() -> None:
    # Drop indexes and constraints created in upgrade
    op.drop_constraint("uq_user_id_role_id", "user_roles", type_="unique")
    op.drop_constraint(
        "uq_scope_permissions_entity_operation_scope_id", "scope_permissions", type_="unique"
    )
    op.drop_index("ix_role_id_entity_type_scope_id", table_name="scope_permissions")
    op.drop_index("ix_id_status", table_name="roles")
    op.drop_constraint(
        "uq_object_permissions_entity_id_operation", "object_permissions", type_="unique"
    )
    op.drop_index("ix_role_id_entity_id", table_name="object_permissions")

    op.rename_table("scope_permissions", "role_permissions")
    op.drop_constraint("pk_scope_permissions", "role_permissions", type_="primary")
    op.create_primary_key("pk_role_permissions", "role_permissions", ["id"])

    op.rename_table("object_permissions", "resource_permissions")
    op.drop_constraint("pk_object_permissions", "resource_permissions", type_="primary")
    op.create_primary_key("pk_resource_permissions", "resource_permissions", ["id"])

    op.add_column(
        "user_roles",
        sa.Column(
            "state",
            sa.VARCHAR(length=16),
            nullable=False,
            server_default="active",
        ),
    )
    op.add_column(
        "user_roles",
        sa.Column("expires_at", sa.DateTime(timezone=True), nullable=True),
    )
    op.add_column(
        "user_roles",
        sa.Column("deleted_at", sa.DateTime(timezone=True), nullable=True),
    )
    op.alter_column("roles", "status", new_column_name="state")
    op.drop_column("role_permissions", "status")
    op.drop_column("resource_permissions", "status")



import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Optional, override

from ai.backend.manager.types import Creator

from .role import RoleData


@dataclass
class UserRoleCreateInput(Creator):
    user_id: uuid.UUID
    role_id: uuid.UUID
    granted_by: Optional[uuid.UUID] = None

    @override
    def fields_to_store(self) -> dict[str, Any]:
        data = {
            "user_id": self.user_id,
            "role_id": self.role_id,
        }
        if self.granted_by is not None:
            data["granted_by"] = self.granted_by
        return data


@dataclass
class UserRoleUpdateInput:
    pass


@dataclass
class UserRoleDeleteInput:
    id: uuid.UUID


@dataclass
class UserRoleData:
    id: uuid.UUID
    user_id: uuid.UUID
    role_id: uuid.UUID
    granted_by: Optional[uuid.UUID]
    granted_at: datetime
    expires_at: Optional[datetime]
    deleted_at: Optional[datetime]


@dataclass
class UserRoleDataWithRole:
    id: uuid.UUID
    user_id: uuid.UUID
    role_id: uuid.UUID
    granted_by: Optional[uuid.UUID]
    granted_at: datetime
    expires_at: Optional[datetime]
    deleted_at: Optional[datetime]

    mapped_role_data: RoleData



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.common.data.storage.registries.types import ModelSortKey
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.artifact.types import ArtifactDataWithRevisions, ArtifactType
from ai.backend.manager.services.artifact.actions.base import ArtifactAction


@dataclass
class ScanArtifactsSyncAction(ArtifactAction):
    artifact_type: Optional[ArtifactType]
    registry_id: Optional[uuid.UUID]
    limit: Optional[int]
    order: Optional[ModelSortKey]
    search: Optional[str]

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "scan_sync"


@dataclass
class ScanArtifactsSyncActionResult(BaseActionResult):
    result: list[ArtifactDataWithRevisions]

    @override
    def entity_id(self) -> Optional[str]:
        return None



import asyncio
import logging
import time
from typing import Awaitable, Callable, ParamSpec, TypeVar

from ai.backend.common.exception import BackendAIError, UnreachableError
from ai.backend.common.metrics.metric import DomainType, LayerMetricObserver, LayerType
from ai.backend.logging import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


P = ParamSpec("P")
R = TypeVar("R")


def create_layer_aware_repository_decorator(
    layer: LayerType,
    default_retry_count: int = 10,
    default_retry_delay: float = 0.1,
):
    """
    Factory function to create layer-aware repository decorators.

    Args:
        layer: The layer type for metric observation
        default_retry_count: Default number of retries for repository operations
        default_retry_delay: Default delay between retries in seconds

    Returns:
        A repository_decorator function configured for the specified layer
    """

    def repository_decorator(
        *,
        retry_count: int = default_retry_count,
        retry_delay: float = default_retry_delay,
    ) -> Callable[
        [Callable[P, Awaitable[R]]],
        Callable[P, Awaitable[R]],
    ]:
        """
        Decorator for repository operations that adds retry logic and metrics.

        Similar to valkey_decorator but for repository layer operations.

        Note: This decorator should only be applied to public methods that are exposed
        to external users. Internal/private methods should not use this decorator.
        """

        def decorator(func: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]:
            observer = LayerMetricObserver.instance()
            operation = func.__name__

            async def wrapper(*args, **kwargs) -> R:
                log.trace("Calling repository method {}", operation)
                start = time.perf_counter()
                for attempt in range(retry_count):
                    try:
                        observer.observe_layer_operation_triggered(
                            domain=DomainType.REPOSITORY,
                            layer=layer,
                            operation=operation,
                        )
                        res = await func(*args, **kwargs)
                        observer.observe_layer_operation(
                            domain=DomainType.REPOSITORY,
                            layer=layer,
                            operation=operation,
                            success=True,
                            duration=time.perf_counter() - start,
                        )
                        return res
                    except BackendAIError as e:
                        log.exception(
                            "Error in repository method {}, args: {}, kwargs: {}, retry_count: {}, error: {}",
                            operation,
                            args,
                            kwargs,
                            retry_count,
                            e,
                        )
                        observer.observe_layer_operation(
                            domain=DomainType.REPOSITORY,
                            layer=layer,
                            operation=operation,
                            success=False,
                            duration=time.perf_counter() - start,
                        )
                        # If it's a BackendAIError, this error is intended to be handled by the caller.
                        raise e
                    except Exception as e:
                        if attempt < retry_count - 1:
                            log.debug(
                                "Retrying repository method {} due to error: {} (attempt {}/{})",
                                operation,
                                e,
                                attempt + 1,
                                retry_count,
                            )
                            observer.observe_layer_retry(
                                domain=DomainType.REPOSITORY,
                                layer=layer,
                                operation=operation,
                            )
                            await asyncio.sleep(retry_delay)
                            continue
                        log.exception(
                            "Error in repository method {}, args: {}, kwargs: {}, retry_count: {}, error: {}",
                            operation,
                            args,
                            kwargs,
                            retry_count,
                            e,
                        )
                        observer.observe_layer_operation(
                            domain=DomainType.REPOSITORY,
                            layer=layer,
                            operation=operation,
                            success=False,
                            duration=time.perf_counter() - start,
                        )
                        raise e
                raise UnreachableError(
                    f"Reached unreachable code in {operation} after {retry_count} attempts"
                )

            return wrapper

        return decorator

    return repository_decorator



from collections.abc import Mapping
from pathlib import Path
from typing import Any, override

from ai.backend.common import config
from ai.backend.manager.config.loader.types import AbstractConfigLoader


class TomlConfigLoader(AbstractConfigLoader):
    _config_path: Path
    _daemon_name: str

    def __init__(self, config_path: Path | str, daemon_name: str) -> None:
        self._config_path = Path(config_path)
        self._daemon_name = daemon_name

    @override
    async def load(self) -> Mapping[str, Any]:
        raw_cfg, _ = config.read_from_file(self._config_path, self._daemon_name)
        return raw_cfg



"""
Hook for interactive session type.
Interactive sessions don't require special handling during state transitions.
"""

import logging

from ai.backend.logging import BraceStyleAdapter

from ..types import SessionTransitionData
from .base import AbstractSessionHook

log = BraceStyleAdapter(logging.getLogger(__name__))


class InteractiveSessionHook(AbstractSessionHook):
    async def on_transition_to_running(self, session: SessionTransitionData) -> None:
        log.debug(
            "Interactive session {} transitioning to RUNNING",
            session.session_id,
        )

    async def on_transition_to_terminated(self, session: SessionTransitionData) -> None:
        log.debug(
            "Interactive session {} transitioning to TERMINATED",
            session.session_id,
        )



from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec
from ai.backend.manager.services.artifact.actions.delegate_scan import (
    DelegateScanArtifactsAction,
    DelegateScanArtifactsActionResult,
)
from ai.backend.manager.services.artifact.actions.delete_multi import (
    DeleteArtifactsAction,
    DeleteArtifactsActionResult,
)
from ai.backend.manager.services.artifact.actions.get import (
    GetArtifactAction,
    GetArtifactActionResult,
)
from ai.backend.manager.services.artifact.actions.get_revisions import (
    GetArtifactRevisionsAction,
    GetArtifactRevisionsActionResult,
)
from ai.backend.manager.services.artifact.actions.list import (
    ListArtifactsAction,
    ListArtifactsActionResult,
)
from ai.backend.manager.services.artifact.actions.list_with_revisions import (
    ListArtifactsWithRevisionsAction,
    ListArtifactsWithRevisionsActionResult,
)
from ai.backend.manager.services.artifact.actions.restore_multi import (
    RestoreArtifactsAction,
    RestoreArtifactsActionResult,
)
from ai.backend.manager.services.artifact.actions.retrieve_model import (
    RetrieveModelAction,
    RetrieveModelActionResult,
)
from ai.backend.manager.services.artifact.actions.retrieve_model_multi import (
    RetrieveModelsAction,
    RetrieveModelsActionResult,
)
from ai.backend.manager.services.artifact.actions.scan import (
    ScanArtifactsAction,
    ScanArtifactsActionResult,
)
from ai.backend.manager.services.artifact.actions.scan_sync import (
    ScanArtifactsSyncAction,
    ScanArtifactsSyncActionResult,
)
from ai.backend.manager.services.artifact.actions.update import (
    UpdateArtifactAction,
    UpdateArtifactActionResult,
)
from ai.backend.manager.services.artifact.actions.upsert_multi import (
    UpsertArtifactsAction,
    UpsertArtifactsActionResult,
)

from .service import ArtifactService


class ArtifactProcessors(AbstractProcessorPackage):
    scan: ActionProcessor[ScanArtifactsAction, ScanArtifactsActionResult]
    scan_sync: ActionProcessor[ScanArtifactsSyncAction, ScanArtifactsSyncActionResult]
    get: ActionProcessor[GetArtifactAction, GetArtifactActionResult]
    list_artifacts: ActionProcessor[ListArtifactsAction, ListArtifactsActionResult]
    list_artifacts_with_revisions: ActionProcessor[
        ListArtifactsWithRevisionsAction, ListArtifactsWithRevisionsActionResult
    ]
    get_revisions: ActionProcessor[GetArtifactRevisionsAction, GetArtifactRevisionsActionResult]
    update: ActionProcessor[UpdateArtifactAction, UpdateArtifactActionResult]
    upsert_artifacts_with_revisions: ActionProcessor[
        UpsertArtifactsAction, UpsertArtifactsActionResult
    ]
    retrieve_models: ActionProcessor[RetrieveModelsAction, RetrieveModelsActionResult]
    retrieve_single_model: ActionProcessor[RetrieveModelAction, RetrieveModelActionResult]
    delete_artifacts: ActionProcessor[DeleteArtifactsAction, DeleteArtifactsActionResult]
    restore_artifacts: ActionProcessor[RestoreArtifactsAction, RestoreArtifactsActionResult]

    delegate_scan: ActionProcessor[DelegateScanArtifactsAction, DelegateScanArtifactsActionResult]

    def __init__(self, service: ArtifactService, action_monitors: list[ActionMonitor]) -> None:
        # TODO: Move scan action to ArtifactRegistryService
        self.scan = ActionProcessor(service.scan, action_monitors)
        self.scan_sync = ActionProcessor(service.scan_sync, action_monitors)
        self.get = ActionProcessor(service.get, action_monitors)
        self.list_artifacts = ActionProcessor(service.list, action_monitors)
        self.list_artifacts_with_revisions = ActionProcessor(
            service.list_with_revisions, action_monitors
        )
        self.get_revisions = ActionProcessor(service.get_revisions, action_monitors)
        self.update = ActionProcessor(service.update, action_monitors)
        self.upsert_artifacts_with_revisions = ActionProcessor(
            service.upsert_artifacts_with_revisions, action_monitors
        )
        self.retrieve_models = ActionProcessor(service.retrieve_models, action_monitors)
        self.retrieve_single_model = ActionProcessor(service.retrieve_single_model, action_monitors)
        self.delete_artifacts = ActionProcessor(service.delete_artifacts, action_monitors)
        self.restore_artifacts = ActionProcessor(service.restore_artifacts, action_monitors)
        self.delegate_scan = ActionProcessor(service.delegate_scan_artifacts, action_monitors)

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            ScanArtifactsAction.spec(),
            ScanArtifactsSyncAction.spec(),
            DelegateScanArtifactsAction.spec(),
            GetArtifactAction.spec(),
            ListArtifactsAction.spec(),
            ListArtifactsWithRevisionsAction.spec(),
            GetArtifactRevisionsAction.spec(),
            UpdateArtifactAction.spec(),
            UpsertArtifactsAction.spec(),
            RetrieveModelAction.spec(),
            RetrieveModelsAction.spec(),
            DeleteArtifactsAction.spec(),
            RestoreArtifactsAction.spec(),
        ]



from dataclasses import dataclass
from typing import Optional, override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.services.permission_contoller.actions.base import RoleAction


@dataclass
class CheckPermissionAction(RoleAction):
    user_id: UUID
    operation: str
    target_entity_type: str
    target_entity_id: str

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.user_id)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "check_permission"


@dataclass
class CheckPermissionActionResult(BaseActionResult):
    has_permission: bool

    @override
    def entity_id(self) -> Optional[str]:
        return None



from __future__ import annotations

import asyncio
import functools
import grp
import importlib
import importlib.resources
import logging
import os
import pwd
import signal
import ssl
import sys
import traceback
from collections.abc import (
    Iterable,
    Mapping,
    MutableMapping,
    Sequence,
)
from contextlib import asynccontextmanager as actxmgr
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from pprint import pformat
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    AsyncIterator,
    Final,
    Optional,
    cast,
)

import aiohttp_cors
import aiomonitor
import aiotools
import click
from aiohttp import web
from aiohttp.typedefs import Handler, Middleware
from setproctitle import setproctitle

from ai.backend.common import redis_helper
from ai.backend.common.auth import PublicKey, SecretKey
from ai.backend.common.bgtask.bgtask import BackgroundTaskManager
from ai.backend.common.cli import LazyGroup
from ai.backend.common.clients.valkey_client.valkey_bgtask.client import ValkeyBgtaskClient
from ai.backend.common.clients.valkey_client.valkey_container_log.client import (
    ValkeyContainerLogClient,
)
from ai.backend.common.clients.valkey_client.valkey_image.client import ValkeyImageClient
from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.clients.valkey_client.valkey_schedule.client import ValkeyScheduleClient
from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.clients.valkey_client.valkey_stream.client import ValkeyStreamClient
from ai.backend.common.config import find_config_file
from ai.backend.common.data.config.types import EtcdConfigData
from ai.backend.common.defs import (
    REDIS_BGTASK_DB,
    REDIS_CONTAINER_LOG,
    REDIS_IMAGE_DB,
    REDIS_LIVE_DB,
    REDIS_STATISTICS_DB,
    REDIS_STREAM_DB,
    REDIS_STREAM_LOCK,
    RedisRole,
)
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.events.dispatcher import EventDispatcher, EventProducer
from ai.backend.common.events.event_types.artifact_registry.anycast import (
    DoScanReservoirRegistryEvent,
)
from ai.backend.common.events.fetcher import EventFetcher
from ai.backend.common.events.hub.hub import EventHub
from ai.backend.common.exception import BackendAIError, ErrorCode
from ai.backend.common.json import dump_json_str
from ai.backend.common.leader.tasks.event_task import EventTaskSpec
from ai.backend.common.message_queue.hiredis_queue import HiRedisQueue
from ai.backend.common.message_queue.queue import AbstractMessageQueue
from ai.backend.common.message_queue.redis_queue import RedisMQArgs, RedisQueue
from ai.backend.common.metrics.http import (
    build_api_metric_middleware,
    build_prometheus_metrics_handler,
)
from ai.backend.common.metrics.metric import CommonMetricRegistry
from ai.backend.common.metrics.profiler import Profiler, PyroscopeArgs
from ai.backend.common.middlewares.request_id import request_id_middleware
from ai.backend.common.msgpack import DEFAULT_PACK_OPTS, DEFAULT_UNPACK_OPTS
from ai.backend.common.plugin.event import EventDispatcherPluginContext
from ai.backend.common.plugin.hook import ALL_COMPLETED, PASSED, HookPluginContext
from ai.backend.common.plugin.monitor import INCREMENT
from ai.backend.common.service_discovery.etcd_discovery.service_discovery import (
    ETCDServiceDiscovery,
    ETCDServiceDiscoveryArgs,
)
from ai.backend.common.service_discovery.redis_discovery.service_discovery import (
    RedisServiceDiscovery,
    RedisServiceDiscoveryArgs,
)
from ai.backend.common.service_discovery.service_discovery import (
    ServiceDiscoveryLoop,
    ServiceEndpoint,
    ServiceMetadata,
)
from ai.backend.common.types import (
    AGENTID_MANAGER,
    AgentSelectionStrategy,
    ServiceDiscoveryType,
)
from ai.backend.common.utils import env_info
from ai.backend.logging import BraceStyleAdapter, Logger, LogLevel
from ai.backend.logging.otel import OpenTelemetrySpec
from ai.backend.manager.config.bootstrap import BootstrapConfig
from ai.backend.manager.config.loader.config_overrider import ConfigOverrider
from ai.backend.manager.config.loader.etcd_loader import (
    EtcdCommonConfigLoader,
    EtcdManagerConfigLoader,
)
from ai.backend.manager.config.loader.legacy_etcd_loader import (
    LegacyEtcdLoader,
    LegacyEtcdVolumesLoader,
)
from ai.backend.manager.config.loader.loader_chain import LoaderChain
from ai.backend.manager.config.loader.toml_loader import TomlConfigLoader
from ai.backend.manager.config.loader.types import AbstractConfigLoader
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.config.watchers.etcd import EtcdConfigWatcher
from ai.backend.manager.sokovan.deployment.deployment_controller import (
    DeploymentController,
    DeploymentControllerArgs,
)
from ai.backend.manager.sokovan.deployment.route.route_controller import (
    RouteController,
    RouteControllerArgs,
)

from . import __version__
from .api.context import RootContext
from .types import DistributedLockFactory, SMTPTriggerPolicy

if TYPE_CHECKING:
    from ai.backend.manager.reporters.base import AbstractReporter

    from .api.types import (
        AppCreator,
        CleanupContext,
        WebRequestHandler,
    )

VALID_VERSIONS: Final = frozenset([
    # 'v1.20160915',  # deprecated
    # 'v2.20170315',  # deprecated
    # 'v3.20170615',  # deprecated
    # authentication changed not to use request bodies
    "v4.20181215",
    # added & enabled streaming-execute API
    "v4.20190115",
    # changed resource/image formats
    "v4.20190315",
    # added user mgmt and ID/password authentication
    # added domain/group/scaling-group
    # added domain/group/scaling-group ref. fields to user/keypair/vfolder objects
    "v4.20190615",
    # added mount_map parameter when creating kernel
    # changed GraphQL query structures for multi-container bundled sessions
    "v5.20191215",
    # rewrote vfolder upload/download APIs to migrate to external storage proxies
    "v6.20200815",
    # added standard-compliant /admin/gql endpoint
    # deprecated /admin/graphql endpoint (still present for backward compatibility)
    # added "groups_by_name" GQL query
    # added "filter" and "order" arg to all paginated GQL queries with their own expression mini-langs
    # removed "order_key" and "order_asc" arguments from all paginated GQL queries (never used!)
    "v6.20210815",
    # added session dependencies and state callback URLs configs when creating sessions
    # added session event webhook option to session creation API
    # added architecture option when making image aliases
    "v6.20220315",
    # added payload encryption / decryption on selected transfer
    "v6.20220615",
    # added config/resource-slots/details, model mgmt & serving APIs
    "v6.20230315",
    # added quota scopes (per-user/per-project quota configs)
    # added user & project resource policies
    # deprecated per-vfolder quota configs (BREAKING)
    "v7.20230615",
    # added /vfolders API set to replace name-based refs to ID-based refs to work with vfolders
    # set pending deprecation for the legacy /folders API set
    # added vfolder trash bin APIs
    # changed the image registry management API to allow per-project registry configs (BREAKING)
    "v8.20240315",
    # added session priority and Relay-compliant ComputeSessioNode, KernelNode queries
    # added dependents/dependees/graph query fields to ComputeSessioNode
    "v8.20240915",
    # added explicit attach_network option to session creation config
    "v9.20250722",
    # added model_ids, model_id_map options to session creation config
    # <future>
    # TODO: replaced keypair-based resource policies to user-based resource policies
    # TODO: began SSO support using per-external-service keypairs (e.g., for FastTrack)
    # TODO: added an initial version of RBAC for projects and vfolders
])
LATEST_REV_DATES: Final = {
    1: "20160915",
    2: "20170915",
    3: "20181215",
    4: "20190615",
    5: "20191215",
    6: "20230315",
    7: "20230615",
    8: "20240915",
    9: "20250722",
}
LATEST_API_VERSION: Final = "v9.20250722"

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

PUBLIC_INTERFACES: Final = [
    "pidx",
    "background_task_manager",
    "db",
    "registry",
    "redis_live",
    "valkey_stat_client",
    "redis_image",
    "redis_stream",
    "event_dispatcher",
    "event_producer",
    "idle_checkers",
    "storage_manager",
    "stats_monitor",
    "error_monitor",
    "hook_plugin_ctx",
]

public_interface_objs: MutableMapping[str, Any] = {}

global_subapp_pkgs: Final[list[str]] = [
    ".acl",
    ".container_registry",
    ".artifact",
    ".artifact_registry",
    ".etcd",
    ".events",
    ".auth",
    ".ratelimit",
    ".vfolder",
    ".admin",
    ".spec",
    ".service",
    ".session",
    ".stream",
    ".manager",
    ".resource",
    ".scaling_group",
    ".cluster_template",
    ".session_template",
    ".image",
    ".userconfig",
    ".domainconfig",
    ".group",
    ".groupconfig",
    ".logs",
    ".object_storage",
]

global_subapp_pkgs_for_public_metrics_app: Final[tuple[str, ...]] = (".health",)

EVENT_DISPATCHER_CONSUMER_GROUP: Final = "manager"


async def hello(request: web.Request) -> web.Response:
    """
    Returns the API version number.
    """
    return web.json_response({
        "version": LATEST_API_VERSION,
        "manager": __version__,
    })


async def on_prepare(request: web.Request, response: web.StreamResponse) -> None:
    response.headers["Server"] = "BackendAI"


@web.middleware
async def api_middleware(request: web.Request, handler: WebRequestHandler) -> web.StreamResponse:
    from .errors.common import GenericBadRequest, InternalServerError

    _handler = handler
    method_override = request.headers.get("X-Method-Override", None)
    if method_override:
        request = request.clone(method=method_override)
        new_match_info = await request.app.router.resolve(request)
        if new_match_info is None:
            raise InternalServerError("No matching method handler found")
        _handler = new_match_info.handler
        request._match_info = new_match_info  # type: ignore  # this is a hack
    ex = request.match_info.http_exception
    if ex is not None:
        # handled by exception_middleware
        raise ex
    new_api_version = request.headers.get("X-BackendAI-Version")
    legacy_api_version = request.headers.get("X-Sorna-Version")
    api_version = new_api_version or legacy_api_version
    try:
        if api_version is None:
            path_major_version = int(request.match_info.get("version", 5))
            revision_date = LATEST_REV_DATES[path_major_version]
            request["api_version"] = (path_major_version, revision_date)
        elif api_version in VALID_VERSIONS:
            hdr_major_version, revision_date = api_version.split(".", maxsplit=1)
            request["api_version"] = (int(hdr_major_version[1:]), revision_date)
        else:
            return GenericBadRequest("Unsupported API version.")
    except (ValueError, KeyError):
        return GenericBadRequest("Unsupported API version.")
    resp = await _handler(request)
    return resp


def _debug_error_response(
    e: Exception,
) -> web.StreamResponse:
    error_type = ""
    error_title = ""
    error_message = "Internal server error"
    status_code = 500
    error_code = ErrorCode.default()
    if isinstance(e, BackendAIError):
        error_type = e.error_type
        error_title = e.error_title
        if e.extra_msg:
            error_message = e.extra_msg
        status_code = e.status_code
        error_code = e.error_code()

    return web.json_response(
        {
            "type": error_type,
            "title": error_title,
            "error_code": str(error_code),
            "msg": error_message,
            "traceback": traceback.format_exc(),
        },
        status=status_code,
        dumps=dump_json_str,
    )


@web.middleware
async def exception_middleware(
    request: web.Request, handler: WebRequestHandler
) -> web.StreamResponse:
    from .errors.api import InvalidAPIParameters
    from .errors.common import (
        GenericBadRequest,
        InternalServerError,
        MethodNotAllowed,
        URLNotFound,
    )
    from .exceptions import InvalidArgument

    root_ctx: RootContext = request.app["_root.context"]
    error_monitor = root_ctx.error_monitor
    stats_monitor = root_ctx.stats_monitor
    method = request.method
    endpoint = getattr(request.match_info.route.resource, "canonical", request.path)
    try:
        await stats_monitor.report_metric(INCREMENT, "ai.backend.manager.api.requests")
        resp = await handler(request)
    # NOTE: pydantic.ValidationError is handled in utils.pydantic_params_api_handler()
    except InvalidArgument as ex:
        if len(ex.args) > 1:
            raise InvalidAPIParameters(f"{ex.args[0]}: {', '.join(map(str, ex.args[1:]))}")
        elif len(ex.args) == 1:
            raise InvalidAPIParameters(ex.args[0])
        else:
            raise InvalidAPIParameters()
    except BackendAIError as ex:
        if ex.status_code // 100 == 4:
            log.warning(
                "client error raised inside handlers: ({} {}): {}", method, endpoint, repr(ex)
            )
        elif ex.status_code // 100 == 5:
            log.exception(
                "Internal server error raised inside handlers: ({} {}): {}",
                method,
                endpoint,
                repr(ex),
            )
        await error_monitor.capture_exception()
        await stats_monitor.report_metric(INCREMENT, "ai.backend.manager.api.failures")
        await stats_monitor.report_metric(
            INCREMENT, f"ai.backend.manager.api.status.{ex.status_code}"
        )
        if root_ctx.config_provider.config.debug.enabled:
            return _debug_error_response(ex)
        raise
    except web.HTTPException as ex:
        await stats_monitor.report_metric(INCREMENT, "ai.backend.manager.api.failures")
        await stats_monitor.report_metric(
            INCREMENT, f"ai.backend.manager.api.status.{ex.status_code}"
        )
        if ex.status_code // 100 == 4:
            log.warning("client error raised inside handlers: ({} {}): {}", method, endpoint, ex)
        elif ex.status_code // 100 == 5:
            log.exception(
                "Internal server error raised inside handlers: ({} {}): {}", method, endpoint, ex
            )
        if ex.status_code == 404:
            raise URLNotFound(extra_data=request.path)
        if ex.status_code == 405:
            concrete_ex = cast(web.HTTPMethodNotAllowed, ex)
            raise MethodNotAllowed(
                method=concrete_ex.method, allowed_methods=concrete_ex.allowed_methods
            )
        raise GenericBadRequest
    except asyncio.CancelledError as e:
        # The server is closing or the client has disconnected in the middle of
        # request.  Atomic requests are still executed to their ends.
        log.debug("Request cancelled ({0} {1})", request.method, request.rel_url)
        raise e
    except Exception as e:
        await error_monitor.capture_exception()
        log.exception(
            "Uncaught exception in HTTP request handlers ({} {}): {}", method, endpoint, e
        )
        if root_ctx.config_provider.config.debug.enabled:
            return _debug_error_response(e)
        else:
            raise InternalServerError()
    else:
        await stats_monitor.report_metric(INCREMENT, f"ai.backend.manager.api.status.{resp.status}")
        return resp


@actxmgr
async def etcd_ctx(root_ctx: RootContext, etcd_config: EtcdConfigData) -> AsyncIterator[None]:
    root_ctx.etcd = AsyncEtcd.initialize(etcd_config)
    yield
    await root_ctx.etcd.close()


@actxmgr
async def config_provider_ctx(
    root_ctx: RootContext,
    log_level: LogLevel,
    config_path: Optional[Path] = None,
    extra_config: Optional[Mapping[str, Any]] = None,
) -> AsyncIterator[ManagerConfigProvider]:
    loaders: list[AbstractConfigLoader] = []

    if config_path:
        toml_config_loader = TomlConfigLoader(config_path, "manager")
        loaders.append(toml_config_loader)
    else:
        log.warning("No config file path specified. Skipped loading toml config file...")

    legacy_etcd_loader = LegacyEtcdLoader(root_ctx.etcd)
    loaders.append(legacy_etcd_loader)
    loaders.append(LegacyEtcdVolumesLoader(root_ctx.etcd))
    loaders.append(EtcdCommonConfigLoader(root_ctx.etcd))
    loaders.append(EtcdManagerConfigLoader(root_ctx.etcd))

    overrides: list[tuple[tuple[str, ...], Any]] = [
        (("debug", "enabled"), log_level == LogLevel.DEBUG),
    ]
    if log_level != LogLevel.NOTSET:
        overrides += [
            (("logging", "level"), log_level),
            (("logging", "pkg-ns", "ai.backend"), log_level),
        ]

    loaders.append(ConfigOverrider(overrides))

    unified_config_loader = LoaderChain(loaders, base_config=extra_config)
    etcd_watcher = EtcdConfigWatcher(root_ctx.etcd)

    config_provider: Optional[ManagerConfigProvider] = None
    try:
        config_provider = await ManagerConfigProvider.create(
            unified_config_loader,
            etcd_watcher,
            legacy_etcd_loader,
        )
        root_ctx.config_provider = config_provider

        if config_provider.config.debug.enabled and root_ctx.pidx == 0:
            print("== Manager configuration ==", file=sys.stderr)
            print(pformat(config_provider.config), file=sys.stderr)
        yield root_ctx.config_provider
    finally:
        if config_provider:
            await config_provider.terminate()


@actxmgr
async def webapp_plugin_ctx(root_app: web.Application) -> AsyncIterator[None]:
    from .plugin.webapp import WebappPluginContext

    root_ctx: RootContext = root_app["_root.context"]
    plugin_ctx = WebappPluginContext(
        root_ctx.etcd,
        root_ctx.config_provider.config.model_dump(by_alias=True),
    )
    await plugin_ctx.init(
        context=root_ctx,
        allowlist=root_ctx.config_provider.config.manager.allowed_plugins,
        blocklist=root_ctx.config_provider.config.manager.disabled_plugins,
    )
    root_ctx.webapp_plugin_ctx = plugin_ctx
    for plugin_name, plugin_instance in plugin_ctx.plugins.items():
        if root_ctx.pidx == 0:
            log.info("Loading webapp plugin: {0}", plugin_name)
        subapp, global_middlewares = await plugin_instance.create_app(root_ctx.cors_options)
        _init_subapp(plugin_name, root_app, subapp, global_middlewares)
    yield
    await plugin_ctx.cleanup()


@actxmgr
async def manager_status_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .api import ManagerStatus

    if root_ctx.pidx == 0:
        mgr_status = await root_ctx.config_provider.legacy_etcd_config_loader.get_manager_status()
        if mgr_status is None or mgr_status not in (ManagerStatus.RUNNING, ManagerStatus.FROZEN):
            # legacy transition: we now have only RUNNING or FROZEN for HA setup.
            await root_ctx.config_provider.legacy_etcd_config_loader.update_manager_status(
                ManagerStatus.RUNNING
            )
            mgr_status = ManagerStatus.RUNNING
        log.info("Manager status: {}", mgr_status)
        tz = root_ctx.config_provider.config.system.timezone
        log.info("Configured timezone: {}", tz.tzname(datetime.now()))
    yield


@actxmgr
async def redis_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    valkey_profile_target = root_ctx.config_provider.config.redis.to_valkey_profile_target()
    root_ctx.valkey_profile_target = valkey_profile_target

    root_ctx.valkey_container_log = await ValkeyContainerLogClient.create(
        valkey_profile_target.profile_target(RedisRole.CONTAINER_LOG),
        db_id=REDIS_CONTAINER_LOG,
        human_readable_name="container_log",  # saving container_log queue
    )
    root_ctx.valkey_live = await ValkeyLiveClient.create(
        valkey_profile_target.profile_target(RedisRole.LIVE),
        db_id=REDIS_LIVE_DB,
        human_readable_name="live",  # tracking live status of various entities
    )
    root_ctx.valkey_stat = await ValkeyStatClient.create(
        valkey_profile_target.profile_target(RedisRole.STATISTICS),
        db_id=REDIS_STATISTICS_DB,
        human_readable_name="stat",  # temporary storage for stat snapshots
    )
    root_ctx.valkey_image = await ValkeyImageClient.create(
        valkey_profile_target.profile_target(RedisRole.IMAGE),
        db_id=REDIS_IMAGE_DB,
        human_readable_name="image",  # per-agent image availability
    )
    root_ctx.valkey_stream = await ValkeyStreamClient.create(
        valkey_profile_target.profile_target(RedisRole.STREAM),
        human_readable_name="stream",
        db_id=REDIS_STREAM_DB,
    )
    root_ctx.valkey_schedule = await ValkeyScheduleClient.create(
        valkey_profile_target.profile_target(RedisRole.STREAM),
        db_id=REDIS_LIVE_DB,
        human_readable_name="schedule",  # scheduling marks and coordination
    )
    root_ctx.valkey_bgtask = await ValkeyBgtaskClient.create(
        valkey_profile_target.profile_target(RedisRole.BGTASK),
        human_readable_name="bgtask",
        db_id=REDIS_BGTASK_DB,
    )
    # Ping ValkeyLiveClient directly
    await root_ctx.valkey_live.get_server_time()
    # ValkeyImageClient has its own connection handling
    # No need to ping it separately as it's already connected
    yield
    await root_ctx.valkey_container_log.close()
    await root_ctx.valkey_image.close()
    await root_ctx.valkey_stat.close()
    await root_ctx.valkey_live.close()
    await root_ctx.valkey_stream.close()
    await root_ctx.valkey_schedule.close()
    await root_ctx.valkey_bgtask.close()


@actxmgr
async def database_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .models.utils import connect_database

    async with connect_database(root_ctx.config_provider.config.db) as db:
        root_ctx.db = db
        yield


def _make_registered_reporters(
    root_ctx: RootContext,
) -> dict[str, AbstractReporter]:
    from .reporters.smtp import SMTPReporter, SMTPSenderArgs

    reporters: dict[str, AbstractReporter] = {}
    smtp_configs = root_ctx.config_provider.config.reporter.smtp
    for smtp_conf in smtp_configs:
        smtp_args = SMTPSenderArgs(
            host=smtp_conf.host,
            port=smtp_conf.port,
            username=smtp_conf.username,
            password=smtp_conf.password,
            sender=smtp_conf.sender,
            recipients=smtp_conf.recipients,
            use_tls=smtp_conf.use_tls,
            max_workers=smtp_conf.max_workers,
            template=smtp_conf.template,
        )
        trigger_policy = SMTPTriggerPolicy[smtp_conf.trigger_policy]
        reporters[smtp_conf.name] = SMTPReporter(smtp_args, trigger_policy)

    return reporters


def _make_action_reporters(
    root_ctx: RootContext,
    reporters: dict[str, AbstractReporter],
) -> dict[str, list[AbstractReporter]]:
    action_monitors: dict[str, list[AbstractReporter]] = {}
    action_monitor_configs = root_ctx.config_provider.config.reporter.action_monitors
    for action_monitor_conf in action_monitor_configs:
        reporter_name: str = action_monitor_conf.reporter
        try:
            reporter = reporters[reporter_name]
        except KeyError:
            log.warning(f'Invalid Reporter: "{reporter_name}"')
            continue

        for action_type in action_monitor_conf.subscribed_actions:
            action_monitors.setdefault(action_type, []).append(reporter)

    return action_monitors


@actxmgr
async def processors_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .actions.monitors.audit_log import AuditLogMonitor
    from .actions.monitors.prometheus import PrometheusMonitor
    from .actions.monitors.reporter import ReporterMonitor
    from .reporters.hub import ReporterHub, ReporterHubArgs
    from .services.processors import ProcessorArgs, Processors, ServiceArgs

    registered_reporters = _make_registered_reporters(root_ctx)
    action_reporters = _make_action_reporters(root_ctx, registered_reporters)
    reporter_hub = ReporterHub(
        ReporterHubArgs(
            reporters=action_reporters,
        )
    )
    reporter_monitor = ReporterMonitor(reporter_hub)
    prometheus_monitor = PrometheusMonitor()
    audit_log_monitor = AuditLogMonitor(root_ctx.db)
    root_ctx.processors = Processors.create(
        ProcessorArgs(
            service_args=ServiceArgs(
                db=root_ctx.db,
                repositories=root_ctx.repositories,
                etcd=root_ctx.etcd,
                config_provider=root_ctx.config_provider,
                storage_manager=root_ctx.storage_manager,
                valkey_stat_client=root_ctx.valkey_stat,
                valkey_live=root_ctx.valkey_live,
                event_fetcher=root_ctx.event_fetcher,
                background_task_manager=root_ctx.background_task_manager,
                event_hub=root_ctx.event_hub,
                agent_registry=root_ctx.registry,
                error_monitor=root_ctx.error_monitor,
                idle_checker_host=root_ctx.idle_checker_host,
                event_dispatcher=root_ctx.event_dispatcher,
                hook_plugin_ctx=root_ctx.hook_plugin_ctx,
                scheduling_controller=root_ctx.scheduling_controller,
                deployment_controller=root_ctx.deployment_controller,
                event_producer=root_ctx.event_producer,
                agent_cache=root_ctx.agent_cache,
            )
        ),
        [reporter_monitor, prometheus_monitor, audit_log_monitor],
    )
    yield


@actxmgr
async def distributed_lock_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    root_ctx.distributed_lock_factory = init_lock_factory(root_ctx)
    yield


@actxmgr
async def event_hub_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    root_ctx.event_hub = EventHub()
    yield
    await root_ctx.event_hub.shutdown()


@actxmgr
async def service_discovery_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    sd_type = root_ctx.config_provider.config.service_discovery.type
    match sd_type:
        case ServiceDiscoveryType.ETCD:
            root_ctx.service_discovery = ETCDServiceDiscovery(
                ETCDServiceDiscoveryArgs(root_ctx.etcd)
            )
        case ServiceDiscoveryType.REDIS:
            live_valkey_target = root_ctx.valkey_profile_target.profile_target(RedisRole.LIVE)
            root_ctx.service_discovery = await RedisServiceDiscovery.create(
                RedisServiceDiscoveryArgs(valkey_target=live_valkey_target)
            )

    root_ctx.sd_loop = ServiceDiscoveryLoop(
        sd_type,
        root_ctx.service_discovery,
        ServiceMetadata(
            display_name=f"manager-{root_ctx.config_provider.config.manager.id}",
            service_group="manager",
            version=__version__,
            endpoint=ServiceEndpoint(
                address=root_ctx.config_provider.config.manager.announce_addr.address,
                port=root_ctx.config_provider.config.manager.announce_addr.port,
                protocol="http",
                prometheus_address=root_ctx.config_provider.config.manager.announce_internal_addr.address,
            ),
        ),
    )

    if root_ctx.config_provider.config.otel.enabled:
        meta = root_ctx.sd_loop.metadata
        otel_spec = OpenTelemetrySpec(
            service_id=meta.id,
            service_name=meta.service_group,
            service_version=meta.version,
            log_level=root_ctx.config_provider.config.otel.log_level,
            endpoint=root_ctx.config_provider.config.otel.endpoint,
        )
        BraceStyleAdapter.apply_otel(otel_spec)
    yield
    root_ctx.sd_loop.close()


@actxmgr
async def message_queue_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    root_ctx.message_queue = await _make_message_queue(root_ctx)
    yield
    await root_ctx.message_queue.close()


@actxmgr
async def event_producer_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    root_ctx.event_fetcher = EventFetcher(root_ctx.message_queue)
    root_ctx.event_producer = EventProducer(
        root_ctx.message_queue,
        source=AGENTID_MANAGER,
        log_events=root_ctx.config_provider.config.debug.log_events,
    )
    yield
    await root_ctx.event_producer.close()
    await asyncio.sleep(0.2)


@actxmgr
async def event_dispatcher_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .event_dispatcher.dispatch import DispatcherArgs, Dispatchers

    root_ctx.event_dispatcher = EventDispatcher(
        root_ctx.message_queue,
        log_events=root_ctx.config_provider.config.debug.log_events,
        event_observer=root_ctx.metrics.event,
    )
    dispatchers = Dispatchers(
        DispatcherArgs(
            root_ctx.valkey_container_log,
            root_ctx.valkey_stat,
            root_ctx.valkey_stream,
            root_ctx.scheduler_dispatcher,
            root_ctx.sokovan_orchestrator.coordinator,
            root_ctx.scheduling_controller,
            root_ctx.sokovan_orchestrator.deployment_coordinator,
            root_ctx.sokovan_orchestrator.route_coordinator,
            root_ctx.repositories.scheduler.repository,
            root_ctx.event_hub,
            root_ctx.registry,
            root_ctx.db,
            root_ctx.idle_checker_host,
            root_ctx.event_dispatcher_plugin_ctx,
            root_ctx.repositories,
            lambda: root_ctx.processors,
            root_ctx.storage_manager,
            root_ctx.config_provider,
            use_sokovan=root_ctx.config_provider.config.manager.use_sokovan,
        )
    )
    dispatchers.dispatch(root_ctx.event_dispatcher)
    await root_ctx.event_dispatcher.start()
    yield
    await root_ctx.event_dispatcher.close()


async def _make_message_queue(
    root_ctx: RootContext,
) -> AbstractMessageQueue:
    redis_profile_target = root_ctx.config_provider.config.redis.to_redis_profile_target()
    stream_redis_target = redis_profile_target.profile_target(RedisRole.STREAM)
    node_id = root_ctx.config_provider.config.manager.id
    args = RedisMQArgs(
        anycast_stream_key="events",
        broadcast_channel="events_all",
        consume_stream_keys={
            "events",
        },
        subscribe_channels={
            "events_all",
        },
        group_name=EVENT_DISPATCHER_CONSUMER_GROUP,
        node_id=node_id,
        db=REDIS_STREAM_DB,
    )
    if root_ctx.config_provider.config.manager.use_experimental_redis_event_dispatcher:
        return HiRedisQueue(
            stream_redis_target,
            args,
        )
    return await RedisQueue.create(
        stream_redis_target,
        args,
    )


@actxmgr
async def idle_checker_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .idle import init_idle_checkers

    root_ctx.idle_checker_host = await init_idle_checkers(
        root_ctx.db,
        root_ctx.config_provider,
        root_ctx.event_producer,
        root_ctx.distributed_lock_factory,
    )
    await root_ctx.idle_checker_host.start()
    yield
    await root_ctx.idle_checker_host.shutdown()


@actxmgr
async def storage_manager_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .models.storage import StorageSessionManager

    root_ctx.storage_manager = StorageSessionManager(root_ctx.config_provider.config.volumes)
    yield
    await root_ctx.storage_manager.aclose()


@actxmgr
async def repositories_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .repositories.repositories import Repositories
    from .repositories.types import RepositoryArgs

    repositories = Repositories.create(
        args=RepositoryArgs(
            db=root_ctx.db,
            storage_manager=root_ctx.storage_manager,
            config_provider=root_ctx.config_provider,
            valkey_stat_client=root_ctx.valkey_stat,
            valkey_live_client=root_ctx.valkey_live,
            valkey_schedule_client=root_ctx.valkey_schedule,
            valkey_image_client=root_ctx.valkey_image,
        )
    )
    root_ctx.repositories = repositories
    yield


@actxmgr
async def network_plugin_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .plugin.network import NetworkPluginContext

    ctx = NetworkPluginContext(
        root_ctx.etcd,
        root_ctx.config_provider.config.model_dump(by_alias=True),
    )
    root_ctx.network_plugin_ctx = ctx
    await ctx.init(
        context=root_ctx,
        allowlist=root_ctx.config_provider.config.manager.allowed_plugins,
        blocklist=root_ctx.config_provider.config.manager.disabled_plugins,
    )
    log.info("NetworkPluginContext initialized with plugins: {}", list(ctx.plugins.keys()))
    yield
    await ctx.cleanup()


@actxmgr
async def hook_plugin_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    ctx = HookPluginContext(
        root_ctx.etcd,
        root_ctx.config_provider.config.model_dump(by_alias=True),
    )
    root_ctx.hook_plugin_ctx = ctx
    await ctx.init(
        context=root_ctx,
        allowlist=root_ctx.config_provider.config.manager.allowed_plugins,
        blocklist=root_ctx.config_provider.config.manager.disabled_plugins,
    )
    hook_result = await ctx.dispatch(
        "ACTIVATE_MANAGER",
        (),
        return_when=ALL_COMPLETED,
    )
    if hook_result.status != PASSED:
        raise RuntimeError("Could not activate the manager instance.")
    yield
    await ctx.cleanup()


@actxmgr
async def event_dispatcher_plugin_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    ctx = EventDispatcherPluginContext(
        root_ctx.etcd,
        root_ctx.config_provider.config.model_dump(by_alias=True),
    )
    root_ctx.event_dispatcher_plugin_ctx = ctx
    await ctx.init(
        context=root_ctx,
        allowlist=root_ctx.config_provider.config.manager.allowed_plugins,
        blocklist=root_ctx.config_provider.config.manager.disabled_plugins,
    )
    yield
    await ctx.cleanup()


@actxmgr
async def agent_registry_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from zmq.auth.certs import load_certificate

    from ai.backend.manager.sokovan.scheduling_controller import (
        SchedulingController,
        SchedulingControllerArgs,
    )

    from .agent_cache import AgentRPCCache
    from .registry import AgentRegistry

    # Create scheduling controller first
    root_ctx.scheduling_controller = SchedulingController(
        SchedulingControllerArgs(
            repository=root_ctx.repositories.scheduler.repository,
            config_provider=root_ctx.config_provider,
            storage_manager=root_ctx.storage_manager,
            event_producer=root_ctx.event_producer,
            valkey_schedule=root_ctx.valkey_schedule,
            network_plugin_ctx=root_ctx.network_plugin_ctx,
            hook_plugin_ctx=root_ctx.hook_plugin_ctx,
        )
    )
    # Create deployment controller
    root_ctx.deployment_controller = DeploymentController(
        DeploymentControllerArgs(
            scheduling_controller=root_ctx.scheduling_controller,
            deployment_repository=root_ctx.repositories.deployment.repository,
            config_provider=root_ctx.config_provider,
            storage_manager=root_ctx.storage_manager,
            event_producer=root_ctx.event_producer,
            valkey_schedule=root_ctx.valkey_schedule,
        )
    )
    root_ctx.route_controller = RouteController(
        RouteControllerArgs(
            valkey_schedule=root_ctx.valkey_schedule,
        )
    )
    manager_pkey, manager_skey = load_certificate(
        root_ctx.config_provider.config.manager.rpc_auth_manager_keypair
    )
    assert manager_skey is not None
    manager_public_key = PublicKey(manager_pkey)
    manager_secret_key = SecretKey(manager_skey)
    root_ctx.agent_cache = AgentRPCCache(root_ctx.db, manager_public_key, manager_secret_key)
    root_ctx.registry = AgentRegistry(
        root_ctx.config_provider,
        root_ctx.db,
        root_ctx.agent_cache,
        root_ctx.valkey_stat,
        root_ctx.valkey_live,
        root_ctx.valkey_image,
        root_ctx.event_producer,
        root_ctx.event_hub,
        root_ctx.storage_manager,
        root_ctx.hook_plugin_ctx,
        root_ctx.network_plugin_ctx,
        root_ctx.scheduling_controller,
        debug=root_ctx.config_provider.config.debug.enabled,
        manager_public_key=manager_public_key,
        manager_secret_key=manager_secret_key,
        use_sokovan=root_ctx.config_provider.config.manager.use_sokovan,
    )
    await root_ctx.registry.init()
    yield
    await root_ctx.registry.shutdown()


@actxmgr
async def sched_dispatcher_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .scheduler.dispatcher import SchedulerDispatcher

    root_ctx.scheduler_dispatcher = await SchedulerDispatcher.create(
        root_ctx.config_provider,
        root_ctx.etcd,
        root_ctx.event_producer,
        root_ctx.distributed_lock_factory,
        root_ctx.registry,
        root_ctx.valkey_live,
        root_ctx.valkey_stat,
        root_ctx.repositories.schedule.repository,
    )
    yield
    await root_ctx.scheduler_dispatcher.close()


@actxmgr
async def leader_election_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    """Initialize leader election for distributed coordination."""
    import socket

    from ai.backend.common.clients.valkey_client.valkey_leader.client import ValkeyLeaderClient
    from ai.backend.common.leader import ValkeyLeaderElection, ValkeyLeaderElectionConfig
    from ai.backend.common.leader.tasks import EventProducerTask, LeaderCron, PeriodicTask

    # Create ValkeyLeaderClient for leader election
    valkey_leader_client = await ValkeyLeaderClient.create(
        valkey_target=root_ctx.valkey_profile_target.profile_target(RedisRole.STREAM),
        db_id=REDIS_STREAM_LOCK,  # Use a dedicated DB for leader election
        human_readable_name="leader",
    )

    # Create leader election configuration
    server_id = f"manager-{socket.gethostname()}-{root_ctx.pidx}"
    leader_config = ValkeyLeaderElectionConfig(
        server_id=server_id,
        leader_key="leader:sokovan:scheduler",
        lease_duration=30,
        renewal_interval=10.0,
        failure_threshold=3,
    )

    # Create leader election instance
    root_ctx.leader_election = ValkeyLeaderElection(
        leader_client=valkey_leader_client,
        config=leader_config,
    )

    # Get task specifications from sokovan and register them
    task_specs = root_ctx.sokovan_orchestrator.create_task_specs()

    # Rescan reservoir registry periodically
    task_specs.append(
        EventTaskSpec(
            name="reservoir_registry_scan",
            event_factory=lambda: DoScanReservoirRegistryEvent(),
            interval=3600,  # 1 hour
            initial_delay=0,
        )
    )

    # Create event producer tasks from specs
    leader_tasks: list[PeriodicTask] = [
        EventProducerTask(spec, root_ctx.event_producer) for spec in task_specs
    ]

    # Register tasks with the election system
    leader_cron = LeaderCron(tasks=leader_tasks)
    root_ctx.leader_election.register_task(leader_cron)

    # Start leader election (will start tasks when becoming leader)
    await root_ctx.leader_election.start()
    log.info(f"Leader election started for server {server_id}")

    yield

    # Cleanup leader election
    await root_ctx.leader_election.stop()


@actxmgr
async def sokovan_orchestrator_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .clients.agent import AgentPool
    from .sokovan.scheduler.factory import create_default_scheduler
    from .sokovan.sokovan import SokovanOrchestrator

    # Create agent pool for scheduler
    agent_pool = AgentPool(root_ctx.agent_cache)

    # Create scheduler with default components
    scheduler = create_default_scheduler(
        root_ctx.repositories.scheduler.repository,
        root_ctx.repositories.deployment.repository,
        root_ctx.config_provider,
        root_ctx.distributed_lock_factory,
        agent_pool,
        root_ctx.network_plugin_ctx,
        root_ctx.event_producer,
        root_ctx.valkey_schedule,
    )

    # Create HTTP client pool for deployment operations
    from ai.backend.common.clients.http_client.client_pool import (
        ClientPool,
        tcp_client_session_factory,
    )
    from ai.backend.manager.sokovan.deployment.coordinator import DeploymentCoordinator
    from ai.backend.manager.sokovan.deployment.route.coordinator import RouteCoordinator

    client_pool = ClientPool(tcp_client_session_factory)

    # Create deployment coordinator
    deployment_coordinator = DeploymentCoordinator(
        valkey_schedule=root_ctx.valkey_schedule,
        deployment_controller=root_ctx.deployment_controller,
        deployment_repository=root_ctx.repositories.deployment.repository,
        event_producer=root_ctx.event_producer,
        lock_factory=root_ctx.distributed_lock_factory,
        config_provider=root_ctx.config_provider,
        scheduling_controller=root_ctx.scheduling_controller,
        client_pool=client_pool,
        valkey_stat=root_ctx.valkey_stat,
        route_controller=root_ctx.route_controller,
    )

    # Create route coordinator
    route_coordinator = RouteCoordinator(
        valkey_schedule=root_ctx.valkey_schedule,
        deployment_repository=root_ctx.repositories.deployment.repository,
        event_producer=root_ctx.event_producer,
        lock_factory=root_ctx.distributed_lock_factory,
        config_provider=root_ctx.config_provider,
        scheduling_controller=root_ctx.scheduling_controller,
        client_pool=client_pool,
    )

    # Create sokovan orchestrator with lock factory for timers
    root_ctx.sokovan_orchestrator = SokovanOrchestrator(
        scheduler=scheduler,
        event_producer=root_ctx.event_producer,
        valkey_schedule=root_ctx.valkey_schedule,
        lock_factory=root_ctx.distributed_lock_factory,
        scheduling_controller=root_ctx.scheduling_controller,
        deployment_coordinator=deployment_coordinator,
        route_coordinator=route_coordinator,
    )

    log.info("Sokovan orchestrator initialized")

    try:
        yield
    finally:
        # Leader election will handle task cleanup
        pass


@actxmgr
async def monitoring_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .plugin.monitor import ManagerErrorPluginContext, ManagerStatsPluginContext

    ectx = ManagerErrorPluginContext(
        root_ctx.etcd, root_ctx.config_provider.config.model_dump(by_alias=True)
    )
    sctx = ManagerStatsPluginContext(
        root_ctx.etcd, root_ctx.config_provider.config.model_dump(by_alias=True)
    )
    init_success = False

    try:
        await ectx.init(
            context={"_root.context": root_ctx},
            allowlist=root_ctx.config_provider.config.manager.allowed_plugins,
        )
        await sctx.init(allowlist=root_ctx.config_provider.config.manager.allowed_plugins)
    except Exception:
        log.error("Failed to initialize monitoring plugins")
    else:
        init_success = True
        root_ctx.error_monitor = ectx
        root_ctx.stats_monitor = sctx
    yield
    if init_success:
        await sctx.cleanup()
        await ectx.cleanup()


@actxmgr
async def services_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    from .service.base import ServicesContext
    from .service.container_registry.base import PerProjectRegistryQuotaRepository
    from .service.container_registry.harbor import (
        PerProjectContainerRegistryQuotaClientPool,
        PerProjectContainerRegistryQuotaService,
    )

    db = root_ctx.db

    per_project_container_registries_quota = PerProjectContainerRegistryQuotaService(
        repository=PerProjectRegistryQuotaRepository(db),
        client_pool=PerProjectContainerRegistryQuotaClientPool(),
    )

    root_ctx.services_ctx = ServicesContext(
        per_project_container_registries_quota,
    )
    yield None


class background_task_ctx:
    def __init__(self, root_ctx: RootContext) -> None:
        self.root_ctx = root_ctx

    async def __aenter__(self) -> None:
        self.root_ctx.background_task_manager = BackgroundTaskManager(
            self.root_ctx.event_producer,
            valkey_client=self.root_ctx.valkey_bgtask,
            server_id=self.root_ctx.config_provider.config.manager.id,
            bgtask_observer=self.root_ctx.metrics.bgtask,
        )

    async def __aexit__(self, *exc_info) -> None:
        pass

    async def shutdown(self) -> None:
        if hasattr(self.root_ctx, "background_task_manager"):
            await self.root_ctx.background_task_manager.shutdown()


def handle_loop_error(
    root_ctx: RootContext,
    loop: asyncio.AbstractEventLoop,
    context: Mapping[str, Any],
) -> None:
    exception = context.get("exception")
    msg = context.get("message", "(empty message)")
    if exception is not None:
        if sys.exc_info()[0] is not None:
            log.exception("Error inside event loop: {0}", msg)
            if (error_monitor := getattr(root_ctx, "error_monitor", None)) is not None:
                loop.create_task(error_monitor.capture_exception())
        else:
            exc_info = (type(exception), exception, exception.__traceback__)
            log.error("Error inside event loop: {0}", msg, exc_info=exc_info)
            if (error_monitor := getattr(root_ctx, "error_monitor", None)) is not None:
                loop.create_task(error_monitor.capture_exception(exc_instance=exception))


def _init_subapp(
    pkg_name: str,
    root_app: web.Application,
    subapp: web.Application,
    global_middlewares: Iterable[Middleware],
) -> None:
    subapp.on_response_prepare.append(on_prepare)

    async def _set_root_ctx(subapp: web.Application):
        # Allow subapp's access to the root app properties.
        # These are the public APIs exposed to plugins as well.
        subapp["_root.context"] = root_app["_root.context"]
        subapp["_root_app"] = root_app

    # We must copy the public interface prior to all user-defined startup signal handlers.
    subapp.on_startup.insert(0, _set_root_ctx)
    if "prefix" not in subapp:
        subapp["prefix"] = pkg_name.split(".")[-1].replace("_", "-")
    prefix = subapp["prefix"]
    root_app.add_subapp("/" + prefix, subapp)
    root_app.middlewares.extend(global_middlewares)


def init_subapp(pkg_name: str, root_app: web.Application, create_subapp: AppCreator) -> None:
    root_ctx: RootContext = root_app["_root.context"]
    subapp, global_middlewares = create_subapp(root_ctx.cors_options)
    _init_subapp(pkg_name, root_app, subapp, global_middlewares)


def init_lock_factory(root_ctx: RootContext) -> DistributedLockFactory:
    ipc_base_path = root_ctx.config_provider.config.manager.ipc_base_path
    manager_id = root_ctx.config_provider.config.manager.id
    lock_backend = root_ctx.config_provider.config.manager.distributed_lock
    log.debug("using {} as the distributed lock backend", lock_backend)
    match lock_backend:
        case "filelock":
            from ai.backend.common.lock import FileLock

            return lambda lock_id, lifetime_hint: FileLock(
                ipc_base_path / f"{manager_id}.{lock_id}.lock",
                timeout=0,
            )
        case "pg_advisory":
            from .pglock import PgAdvisoryLock

            return lambda lock_id, lifetime_hint: PgAdvisoryLock(root_ctx.db, lock_id)
        case "redlock":
            from ai.backend.common.lock import RedisLock

            redlock_config = root_ctx.config_provider.config.manager.redlock_config
            redis_profile_target = root_ctx.config_provider.config.redis.to_redis_profile_target()
            redis_lock = redis_helper.get_redis_object(
                redis_profile_target.profile_target(RedisRole.STREAM_LOCK),
                name="lock",  # distributed locks
                db=REDIS_STREAM_LOCK,
            )
            return lambda lock_id, lifetime_hint: RedisLock(
                str(lock_id),
                redis_lock,
                lifetime=min(lifetime_hint * 2, lifetime_hint + 30),
                lock_retry_interval=redlock_config["lock_retry_interval"],
            )
        case "etcd":
            from ai.backend.common.lock import EtcdLock

            return lambda lock_id, lifetime_hint: EtcdLock(
                str(lock_id),
                root_ctx.etcd,
                lifetime=min(lifetime_hint * 2, lifetime_hint + 30),
            )
        case other:
            raise ValueError(f"Invalid lock backend: {other}")


def build_root_app(
    pidx: int,
    bootstrap_config: BootstrapConfig,
    *,
    cleanup_contexts: Optional[Sequence[CleanupContext]] = None,
    subapp_pkgs: Optional[Sequence[str]] = None,
    scheduler_opts: Optional[Mapping[str, Any]] = None,
) -> web.Application:
    from .sweeper.kernel import stale_kernel_sweeper_ctx
    from .sweeper.session import stale_session_sweeper_ctx

    public_interface_objs.clear()
    if bootstrap_config.pyroscope.enabled:
        if (
            not bootstrap_config.pyroscope.app_name
            or not bootstrap_config.pyroscope.server_addr
            or not bootstrap_config.pyroscope.sample_rate
        ):
            raise ValueError("Pyroscope configuration is incomplete.")

        Profiler(
            pyroscope_args=PyroscopeArgs(
                enabled=bootstrap_config.pyroscope.enabled,
                application_name=bootstrap_config.pyroscope.app_name,
                server_address=bootstrap_config.pyroscope.server_addr,
                sample_rate=bootstrap_config.pyroscope.sample_rate,
            )
        )

    root_ctx = RootContext()
    root_ctx.metrics = CommonMetricRegistry.instance()
    app = web.Application(
        middlewares=[
            request_id_middleware,
            exception_middleware,
            api_middleware,
            build_api_metric_middleware(root_ctx.metrics.api),
        ]
    )
    global_exception_handler = functools.partial(handle_loop_error, root_ctx)
    loop = asyncio.get_running_loop()
    loop.set_exception_handler(global_exception_handler)
    app["_root.context"] = root_ctx

    # If the request path starts with the following route, the auth_middleware is bypassed.
    # In this case, all authentication flags are turned off.
    # Used in special cases where the request headers cannot be modified.
    app["auth_middleware_allowlist"] = [
        "/container-registries/webhook",
    ]

    root_ctx.pidx = pidx
    root_ctx.cors_options = {
        "*": aiohttp_cors.ResourceOptions(
            allow_credentials=False, expose_headers="*", allow_headers="*"
        ),
    }
    default_scheduler_opts = {
        "limit": 2048,
        "close_timeout": 30,
        "exception_handler": global_exception_handler,
        "agent_selection_strategy": AgentSelectionStrategy.DISPERSED,
    }
    app["scheduler_opts"] = {
        **default_scheduler_opts,
        **(scheduler_opts if scheduler_opts is not None else {}),
    }
    app.on_response_prepare.append(on_prepare)

    if cleanup_contexts is None:
        cleanup_contexts = [
            event_hub_ctx,
            manager_status_ctx,
            redis_ctx,
            database_ctx,
            services_ctx,
            distributed_lock_ctx,
            message_queue_ctx,
            event_producer_ctx,
            storage_manager_ctx,
            repositories_ctx,
            hook_plugin_ctx,
            monitoring_ctx,
            network_plugin_ctx,
            event_dispatcher_plugin_ctx,
            idle_checker_ctx,
            agent_registry_ctx,
            sched_dispatcher_ctx,
            sokovan_orchestrator_ctx,
            leader_election_ctx,
            event_dispatcher_ctx,
            background_task_ctx,
            stale_session_sweeper_ctx,
            stale_kernel_sweeper_ctx,
            processors_ctx,
            service_discovery_ctx,
        ]

    async def _cleanup_context_wrapper(cctx, app: web.Application) -> AsyncIterator[None]:
        # aiohttp's cleanup contexts are just async generators, not async context managers.
        cctx_instance = cctx(app["_root.context"])
        app["_cctx_instances"].append(cctx_instance)
        try:
            async with cctx_instance:
                yield
        except Exception as e:
            exc_info = (type(e), e, e.__traceback__)
            log.error("Error initializing cleanup_contexts: {0}", cctx.__name__, exc_info=exc_info)

    async def _call_cleanup_context_shutdown_handlers(app: web.Application) -> None:
        for cctx in app["_cctx_instances"]:
            if hasattr(cctx, "shutdown"):
                try:
                    await cctx.shutdown()
                except Exception:
                    log.exception("error while shutting down a cleanup context")

    app["_cctx_instances"] = []
    app.on_shutdown.append(_call_cleanup_context_shutdown_handlers)
    for cleanup_ctx in cleanup_contexts:
        app.cleanup_ctx.append(
            functools.partial(_cleanup_context_wrapper, cleanup_ctx),
        )
    cors = aiohttp_cors.setup(app, defaults=root_ctx.cors_options)
    # should be done in create_app() in other modules.
    cors.add(app.router.add_route("GET", r"", hello))
    cors.add(app.router.add_route("GET", r"/", hello))
    if subapp_pkgs is None:
        subapp_pkgs = []
    for pkg_name in subapp_pkgs:
        if pidx == 0:
            log.info("Loading module: {0}", pkg_name[1:])
        subapp_mod = importlib.import_module(pkg_name, "ai.backend.manager.api")
        init_subapp(pkg_name, app, getattr(subapp_mod, "create_app"))

    vendor_path = importlib.resources.files("ai.backend.manager.vendor")
    assert isinstance(vendor_path, Path)
    app.router.add_static("/static/vendor", path=vendor_path, name="static")
    return app


def build_prometheus_service_discovery_handler(
    root_ctx: RootContext,
) -> Handler:
    async def _handler(request: web.Request) -> web.Response:
        services = await root_ctx.service_discovery.discover()
        resp = []
        for service in services:
            resp.append({
                "targets": [f"{service.endpoint.prometheus_address}"],
                "labels": {
                    "service_id": service.id,
                    "service_group": service.service_group,
                    "display_name": service.display_name,
                    "version": service.version,
                },
            })

        return web.json_response(
            resp,
            status=200,
            dumps=dump_json_str,
        )

    return _handler


def build_internal_app(root_ctx: RootContext) -> web.Application:
    app = web.Application()
    metric_registry = CommonMetricRegistry.instance()
    app.router.add_route("GET", r"/metrics", build_prometheus_metrics_handler(metric_registry))
    app.router.add_route(
        "GET", r"/metrics/service_discovery", build_prometheus_service_discovery_handler(root_ctx)
    )
    return app


def build_public_app(
    root_ctx: RootContext,
    subapp_pkgs: Iterable[str] | None = None,
) -> web.Application:
    app = web.Application()
    app["_root.context"] = root_ctx
    if subapp_pkgs is None:
        subapp_pkgs = []
    for pkg_name in subapp_pkgs:
        if root_ctx.pidx == 0:
            log.info("Loading module: {0}", pkg_name[1:])
        subapp_mod = importlib.import_module(pkg_name, "ai.backend.manager.public_api")
        init_subapp(pkg_name, app, getattr(subapp_mod, "create_app"))
    return app


@dataclass
class ServerMainArgs:
    bootstrap_cfg: BootstrapConfig
    bootstrap_cfg_path: Path
    log_endpoint: str
    log_level: LogLevel


@actxmgr
async def server_main(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    args: ServerMainArgs,
) -> AsyncIterator[None]:
    boostrap_config = args.bootstrap_cfg

    root_app = build_root_app(pidx, boostrap_config, subapp_pkgs=global_subapp_pkgs)
    root_ctx: RootContext = root_app["_root.context"]
    internal_app = build_internal_app(root_ctx)

    # Start aiomonitor.
    # Port is set by config (default=50100 + pidx).
    loop.set_debug(boostrap_config.debug.asyncio)
    m = aiomonitor.Monitor(
        loop,
        termui_port=boostrap_config.manager.aiomonitor_termui_port + pidx,
        webui_port=boostrap_config.manager.aiomonitor_webui_port + pidx,
        console_enabled=False,
        hook_task_factory=boostrap_config.debug.enhanced_aiomonitor_task_info,
    )
    m.prompt = f"monitor (manager[{pidx}@{os.getpid()}]) >>> "
    # Add some useful console_locals for ease of debugging
    m.console_locals["root_app"] = root_app
    m.console_locals["root_ctx"] = root_ctx
    aiomon_started = False
    try:
        m.start()
        aiomon_started = True
    except Exception as e:
        log.warning("aiomonitor could not start but skipping this error to continue", exc_info=e)

    # Plugin webapps should be loaded before runner.setup(),
    # which freezes on_startup event.
    try:
        async with (
            etcd_ctx(root_ctx, boostrap_config.etcd.to_dataclass()),
            config_provider_ctx(root_ctx, args.log_level, args.bootstrap_cfg_path),
            webapp_plugin_ctx(root_app),
        ):
            ssl_ctx = None
            if root_ctx.config_provider.config.manager.ssl_enabled:
                ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
                ssl_ctx.load_cert_chain(
                    str(root_ctx.config_provider.config.manager.ssl_cert),
                    root_ctx.config_provider.config.manager.ssl_privkey,
                )

            runner = web.AppRunner(root_app, keepalive_timeout=30.0)
            internal_runner = web.AppRunner(internal_app, keepalive_timeout=30.0)
            await runner.setup()
            await internal_runner.setup()
            service_addr = root_ctx.config_provider.config.manager.service_addr
            internal_addr = root_ctx.config_provider.config.manager.internal_addr
            site = web.TCPSite(
                runner,
                service_addr.host,
                service_addr.port,
                backlog=1024,
                reuse_port=True,
                ssl_context=ssl_ctx,
            )
            internal_site = web.TCPSite(
                internal_runner,
                internal_addr.host,
                internal_addr.port,
                backlog=1024,
                reuse_port=True,
            )
            await site.start()
            await internal_site.start()
            public_metrics_port = root_ctx.config_provider.config.manager.public_metrics_port
            if public_metrics_port is not None:
                _app = build_public_app(
                    root_ctx, subapp_pkgs=global_subapp_pkgs_for_public_metrics_app
                )
                _runner = web.AppRunner(_app, keepalive_timeout=30.0)
                await _runner.setup()
                _site = web.TCPSite(
                    _runner,
                    service_addr.host,
                    public_metrics_port,
                    backlog=1024,
                    reuse_port=True,
                )
                await _site.start()
                log.info(
                    f"started handling public metric API requests at {service_addr.host}:{public_metrics_port}"
                )

            if os.geteuid() == 0:
                uid = root_ctx.config_provider.config.manager.user
                gid = root_ctx.config_provider.config.manager.group
                if uid is None or gid is None:
                    raise ValueError("user/group must be specified when running as root")

                os.setgroups([
                    g.gr_gid for g in grp.getgrall() if pwd.getpwuid(uid).pw_name in g.gr_mem
                ])
                os.setgid(gid)
                os.setuid(uid)
                log.info("changed process uid and gid to {}:{}", uid, gid)
            log.info("started handling API requests at {}", service_addr)

            try:
                yield
            finally:
                log.info("shutting down...")
                await runner.cleanup()
    finally:
        if aiomon_started:
            m.close()


@aiotools.server_context
async def server_main_logwrapper(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    tuple_args: Sequence[Any],
) -> AsyncGenerator[None, signal.Signals]:
    setproctitle(f"backend.ai: manager worker-{pidx}")
    args = ServerMainArgs(
        bootstrap_cfg=tuple_args[0],
        bootstrap_cfg_path=tuple_args[1],
        log_endpoint=tuple_args[2],
        log_level=tuple_args[3],
    )
    logger = Logger(
        args.bootstrap_cfg.logging,
        is_master=False,
        log_endpoint=args.log_endpoint,
        msgpack_options={
            "pack_opts": DEFAULT_PACK_OPTS,
            "unpack_opts": DEFAULT_UNPACK_OPTS,
        },
    )
    try:
        with logger:
            async with server_main(loop, pidx, args):
                yield
    except Exception:
        traceback.print_exc()


@click.group(invoke_without_command=True)
@click.option(
    "-f",
    "--config-path",
    "--config",
    type=Path,
    default=None,
    help="The config file path. (default: ./manager.toml and /etc/backend.ai/manager.toml)",
)
@click.option(
    "--debug",
    is_flag=True,
    help="A shortcut to set `--log-level=DEBUG`",
)
@click.option(
    "--log-level",
    type=click.Choice([*LogLevel], case_sensitive=False),
    default=LogLevel.NOTSET,
    help="Set the logging verbosity level",
)
@click.pass_context
def main(
    ctx: click.Context,
    config_path: Optional[Path],
    debug: bool,
    log_level: LogLevel,
) -> None:
    """
    Start the manager service as a foreground process.
    """
    log_level = LogLevel.DEBUG if debug else log_level

    if config_path is None:
        discovered_cfg_path = find_config_file("manager")
    else:
        discovered_cfg_path = Path(config_path)

    bootstrap_cfg = asyncio.run(BootstrapConfig.load_from_file(discovered_cfg_path, log_level))

    if ctx.invoked_subcommand is None:
        bootstrap_cfg.manager.pid_file.write_text(str(os.getpid()))
        ipc_base_path = bootstrap_cfg.manager.ipc_base_path
        log_sockpath = ipc_base_path / f"manager-logger-{os.getpid()}.sock"
        log_endpoint = f"ipc://{log_sockpath}"
        try:
            logger = Logger(
                bootstrap_cfg.logging,
                is_master=True,
                log_endpoint=log_endpoint,
                msgpack_options={
                    "pack_opts": DEFAULT_PACK_OPTS,
                    "unpack_opts": DEFAULT_UNPACK_OPTS,
                },
            )
            with logger:
                ns = bootstrap_cfg.etcd.namespace
                setproctitle(f"backend.ai: manager {ns}")
                log.info("Backend.AI Manager {0}", __version__)
                log.info("runtime: {0}", env_info())
                log_config = logging.getLogger("ai.backend.manager.config")
                log_config.debug("debug mode enabled.")
                if bootstrap_cfg.manager.event_loop == "uvloop":
                    import uvloop

                    uvloop.install()
                    log.info("Using uvloop as the event loop backend")
                try:
                    aiotools.start_server(
                        server_main_logwrapper,
                        num_workers=bootstrap_cfg.manager.num_proc,
                        args=(bootstrap_cfg, discovered_cfg_path, log_endpoint, log_level),
                        wait_timeout=5.0,
                    )
                finally:
                    log.info("terminated.")
        finally:
            if bootstrap_cfg.manager.pid_file.is_file():
                # check is_file() to prevent deleting /dev/null!
                bootstrap_cfg.manager.pid_file.unlink()
    else:
        # Click is going to invoke a subcommand.
        pass


@main.group(cls=LazyGroup, import_name="ai.backend.manager.api.auth:cli")
def auth() -> None:
    pass


if __name__ == "__main__":
    sys.exit(main())



import uuid

import sqlalchemy as sa
from sqlalchemy.orm import selectinload

from ai.backend.common.data.artifact.types import ArtifactRegistryType
from ai.backend.common.metrics.metric import LayerType
from ai.backend.manager.data.artifact_registries.types import (
    ArtifactRegistryCreatorMeta,
    ArtifactRegistryModifierMeta,
)
from ai.backend.manager.data.huggingface_registry.creator import HuggingFaceRegistryCreator
from ai.backend.manager.data.huggingface_registry.modifier import HuggingFaceRegistryModifier
from ai.backend.manager.data.huggingface_registry.types import HuggingFaceRegistryData
from ai.backend.manager.decorators.repository_decorator import (
    create_layer_aware_repository_decorator,
)
from ai.backend.manager.errors.artifact import ArtifactNotFoundError
from ai.backend.manager.errors.artifact_registry import ArtifactRegistryNotFoundError
from ai.backend.manager.models.artifact import ArtifactRow
from ai.backend.manager.models.artifact_registries import ArtifactRegistryRow
from ai.backend.manager.models.huggingface_registry import HuggingFaceRegistryRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

# Layer-specific decorator for artifact registry repository
repository_decorator = create_layer_aware_repository_decorator(LayerType.ARTIFACT_REGISTRY)


class HuggingFaceRepository:
    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    @repository_decorator()
    async def get_registry_data_by_id(self, registry_id: uuid.UUID) -> HuggingFaceRegistryData:
        async with self._db.begin_session() as db_sess:
            result = await db_sess.execute(
                sa.select(HuggingFaceRegistryRow)
                .where(HuggingFaceRegistryRow.id == registry_id)
                .options(selectinload(HuggingFaceRegistryRow.meta))
            )
            row: HuggingFaceRegistryRow = result.scalar_one_or_none()
            if row is None:
                raise ArtifactRegistryNotFoundError(f"Registry with ID {registry_id} not found")
            return row.to_dataclass()

    @repository_decorator()
    async def get_registry_data_by_name(self, name: str) -> HuggingFaceRegistryData:
        async with self._db.begin_session() as db_sess:
            result = await db_sess.execute(
                sa.select(ArtifactRegistryRow)
                .where(ArtifactRegistryRow.name == name)
                .options(
                    selectinload(ArtifactRegistryRow.huggingface_registries).selectinload(
                        HuggingFaceRegistryRow.meta
                    )
                )
            )
            row: ArtifactRegistryRow = result.scalar_one_or_none()
            if row is None:
                raise ArtifactRegistryNotFoundError(f"Registry with name {name} not found")
            return row.huggingface_registries.to_dataclass()

    @repository_decorator()
    async def get_registry_data_by_artifact_id(
        self, artifact_id: uuid.UUID
    ) -> HuggingFaceRegistryData:
        async with self._db.begin_session() as db_sess:
            result = await db_sess.execute(
                sa.select(ArtifactRow)
                .where(ArtifactRow.id == artifact_id)
                .options(
                    selectinload(ArtifactRow.huggingface_registry).selectinload(
                        HuggingFaceRegistryRow.meta
                    ),
                )
            )
            row: ArtifactRow = result.scalar_one_or_none()
            if row is None:
                raise ArtifactNotFoundError(f"Artifact with ID {artifact_id} not found")
            return row.huggingface_registry.to_dataclass()

    @repository_decorator()
    async def create(
        self, creator: HuggingFaceRegistryCreator, meta: ArtifactRegistryCreatorMeta
    ) -> HuggingFaceRegistryData:
        """
        Create a new Hugging Face registry entry.
        """
        async with self._db.begin_session() as db:
            hf_insert = (
                sa.insert(HuggingFaceRegistryRow)
                .values(**creator.fields_to_store())
                .returning(HuggingFaceRegistryRow.id)
            )
            hf_id = (await db.execute(hf_insert)).scalar_one()

            reg_insert = sa.insert(ArtifactRegistryRow).values(
                name=meta.name,
                registry_id=hf_id,
                type=ArtifactRegistryType.HUGGINGFACE,
            )
            await db.execute(reg_insert)

            stmt = (
                sa.select(HuggingFaceRegistryRow)
                .where(HuggingFaceRegistryRow.id == hf_id)
                .options(selectinload(HuggingFaceRegistryRow.meta))
            )
            row: HuggingFaceRegistryRow | None = (await db.execute(stmt)).scalar_one_or_none()
            if row is None:
                raise ArtifactRegistryNotFoundError(f"Registry with ID {hf_id} not found")

            return row.to_dataclass()

    @repository_decorator()
    async def update(
        self,
        registry_id: uuid.UUID,
        modifier: HuggingFaceRegistryModifier,
        meta: ArtifactRegistryModifierMeta,
    ) -> HuggingFaceRegistryData:
        """
        Update an existing Hugging Face registry entry in the database.
        """
        async with self._db.begin_session() as db_session:
            data = modifier.fields_to_update()

            update_stmt = (
                sa.update(HuggingFaceRegistryRow)
                .where(HuggingFaceRegistryRow.id == registry_id)
                .values(**data)
                .returning(HuggingFaceRegistryRow.id)
            )

            result = await db_session.execute(update_stmt)
            inserted_row_id = result.scalar()

            if (name := meta.name.optional_value()) is not None:
                await db_session.execute(
                    sa.update(ArtifactRegistryRow)
                    .where(ArtifactRegistryRow.registry_id == inserted_row_id)
                    .values(name=name)
                )

            # Reselect for the `selectinload`
            row = (
                await db_session.execute(
                    sa.select(HuggingFaceRegistryRow)
                    .where(HuggingFaceRegistryRow.id == inserted_row_id)
                    .options(selectinload(HuggingFaceRegistryRow.meta))
                )
            ).scalar_one()

            return row.to_dataclass()

    @repository_decorator()
    async def delete(self, registry_id: uuid.UUID) -> uuid.UUID:
        """
        Delete an existing Hugging Face registry entry from the database.
        """
        async with self._db.begin_session() as db_session:
            delete_query = (
                sa.delete(HuggingFaceRegistryRow)
                .where(HuggingFaceRegistryRow.id == registry_id)
                .returning(HuggingFaceRegistryRow.id)
            )
            result = await db_session.execute(delete_query)
            deleted_id = result.scalar()

            delete_meta_query = sa.delete(ArtifactRegistryRow).where(
                ArtifactRegistryRow.id == registry_id
            )
            await db_session.execute(delete_meta_query)
            return deleted_id

    @repository_decorator()
    async def get_registries_by_ids(
        self, registry_ids: list[uuid.UUID]
    ) -> list[HuggingFaceRegistryData]:
        """
        Get multiple Hugging Face registry entries by their IDs in a single query.
        """
        async with self._db.begin_session() as db_session:
            result = await db_session.execute(
                sa.select(HuggingFaceRegistryRow)
                .where(HuggingFaceRegistryRow.id.in_(registry_ids))
                .options(selectinload(HuggingFaceRegistryRow.meta))
            )
            rows: list[HuggingFaceRegistryRow] = result.scalars().all()
            return [row.to_dataclass() for row in rows]

    @repository_decorator()
    async def list_registries(self) -> list[HuggingFaceRegistryData]:
        """
        List all Hugging Face registry entries from the database.
        """
        async with self._db.begin_session() as db_session:
            query = sa.select(HuggingFaceRegistryRow).options(
                selectinload(HuggingFaceRegistryRow.meta)
            )
            result = await db_session.execute(query)
            rows: list[HuggingFaceRegistryRow] = result.scalars().all()
            return [row.to_dataclass() for row in rows]



import logging

from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.repositories.user_resource_policy.repository import (
    UserResourcePolicyRepository,
)
from ai.backend.manager.services.user_resource_policy.actions.create_user_resource_policy import (
    CreateUserResourcePolicyAction,
    CreateUserResourcePolicyActionResult,
)
from ai.backend.manager.services.user_resource_policy.actions.delete_user_resource_policy import (
    DeleteUserResourcePolicyAction,
    DeleteUserResourcePolicyActionResult,
)
from ai.backend.manager.services.user_resource_policy.actions.modify_user_resource_policy import (
    ModifyUserResourcePolicyAction,
    ModifyUserResourcePolicyActionResult,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class UserResourcePolicyService:
    _user_resource_policy_repository: UserResourcePolicyRepository

    def __init__(
        self,
        user_resource_policy_repository: UserResourcePolicyRepository,
    ) -> None:
        self._user_resource_policy_repository = user_resource_policy_repository

    async def create_user_resource_policy(
        self, action: CreateUserResourcePolicyAction
    ) -> CreateUserResourcePolicyActionResult:
        creator = action.creator
        to_create = creator.fields_to_store()
        result = await self._user_resource_policy_repository.create(to_create)
        return CreateUserResourcePolicyActionResult(user_resource_policy=result)

    async def modify_user_resource_policy(
        self, action: ModifyUserResourcePolicyAction
    ) -> ModifyUserResourcePolicyActionResult:
        name = action.name
        modifier = action.modifier
        to_update = modifier.fields_to_update()
        result = await self._user_resource_policy_repository.update(name, to_update)
        return ModifyUserResourcePolicyActionResult(user_resource_policy=result)

    async def delete_user_resource_policy(
        self, action: DeleteUserResourcePolicyAction
    ) -> DeleteUserResourcePolicyActionResult:
        name = action.name
        result = await self._user_resource_policy_repository.delete(name)
        return DeleteUserResourcePolicyActionResult(user_resource_policy=result)
